from __future__ import annotations
from .intervals import Interval as I
from .pbox_abc import Pbox, Staircase, pbox_from_ecdf_bundle
from .utils import NotIncreasingError
from typing import *
from warnings import warn
import numpy as np
import matplotlib.pyplot as plt
from .params import Params
from .ecdf import pl_ecdf_bounding_bundles, get_ecdf, eCDF_bundle
from .imprecise import imprecise_ecdf
from numbers import Number
from ..decorator import exposeUN
from numpy.typing import ArrayLike


""" non-parametric pbox  """


__all__ = [
    "known_properties",
    "known_constraints",
    "min_max",
    "min_max_mean",
    "min_mean",
    "min_max_mean_std",
    "min_max_mean_var",
    "min_max_mode",
    "min_max_median",
    # "min_max_median_is_mode",
    "mean_std",
    "mean_var",
    "pos_mean_std",
    # "symmetric_mean_std",
    "from_percentiles",
    "KS_bounds",
]
# ---------------------from data---------------------#

if TYPE_CHECKING:
    from .utils import eCDF_bundle
    from ..characterisation.uncertainNumber import UncertainNumber
    from .pbox_abc import Pbox


def KS_bounds(
    s: ArrayLike, alpha: float, display=True, output_type="bounds"
) -> Union[tuple[eCDF_bundle], Pbox, UncertainNumber]:
    """construct free pbox from sample data by Kolmogorov-Smirnoff confidence bounds

    args:
        s (ArrayLike): sample data, precise and imprecise

        dn (float): KS critical value at a significance level and sample size N;

        output_type (str): A choice between {'bounds', 'pbox', 'un'}, default='bounds'
            which returns two eCDF bundles as bounds; 'pbox' to return a pbox object; 'un' to return an uncertain number object.

    return:
        a tuple of two CDF bounds, i.e. upper and lower (eCDF_bundle objects), or a Pbox object, or an UncertainNumber object
        the return type is controlled by the `output_type` argument.

    note:
        By default the function returns two eCDF bundles as the extreme bounds. With the upper and lower bounds, a free pbox can be constructed.

    example:
        >>> # both precise data (e.g. numpy array) and imprecise data (e.g. a vector of interval) are supported
        >>> precise_data = np.random.normal(0, 1, 100)  # precise data case
        >>> ub, lb = pba.KS_bounds(precise_data, alpha=0.025, display=True)

        >>> # alternatively, an uncertain number or a p-box can be returned
        >>> pba.KS_bounds(precise_data, alpha=0.025, display=False, output_type='pbox')  # return a pbox object
        >>> pba.KS_bounds(precise_data, alpha=0.025, display=False, output_type='un')  # return an uncertain number object

        >>> # imprecise data case
        >>> impre_data = pba.I(lo = precise_data -0.5, hi = precise_data + 0.5)
        >>> ub, lb = pba.KS_bounds(impre_data, alpha=0.025, display=True)


    .. figure:: /_static/ks_bounds_demo.png
        :alt: ks_bounds
        :align: center
        :width: 50%

        Kolmogorov-Smirnoff confidence bounds illustration with precise and imprecise data.
    """

    from pyuncertainnumber import UncertainNumber

    def inner(s, alpha, display):
        dn = d_alpha(len(s), alpha)
        # precise data
        if isinstance(s, list | np.ndarray):
            # ecdf = sps.ecdf(s)
            # b = transform_eeCDF_bundle(ecdf)

            q, p = get_ecdf(s)
            f_l, f_r = p + dn, p - dn
            f_l, f_r = logical_bounding(f_l), logical_bounding(f_r)
            # new ecdf bundles
            b_l, b_r = eCDF_bundle(q, f_l), eCDF_bundle(q, f_r)

            if display:
                fig, ax = plt.subplots()
                ax.step(q, p, color="black", ls=":", where="post")
                pl_ecdf_bounding_bundles(b_l, b_r, ax=ax)
            return b_l, b_r
        # imprecise data
        elif isinstance(s, I):
            b_l, b_r = imprecise_ecdf(s)
            b_lbp, b_rbp = imprecise_ecdf(s)

            b_l.probabilities += dn
            b_r.probabilities -= dn

            b_l.probabilities, b_r.probabilities = logical_bounding(
                b_l.probabilities
            ), logical_bounding(b_r.probabilities)

            if display:
                fig, ax = plt.subplots()
                # plot the epimirical ecdf
                ax.plot(
                    b_lbp.quantiles,
                    b_lbp.probabilities,
                    drawstyle="steps-post",
                    ls=":",
                    color="gray",
                )
                ax.plot(
                    b_rbp.quantiles,
                    b_rbp.probabilities,
                    drawstyle="steps-post",
                    ls=":",
                    color="gray",
                )

                # plot the KS bounds
                pl_ecdf_bounding_bundles(
                    b_l,
                    b_r,
                    sig_level=(1 - 2 * alpha) * 100,
                    ax=ax,
                    title=f"Kolmogorov-Smirnoff confidence bounds at {(1 - 2 * alpha) * 100} % confidence level",
                )
        else:
            raise ValueError("Invalid input data type")
        return b_l, b_r

    b_l, b_r = inner(s, alpha, display)
    if output_type == "bounds":
        return b_l, b_r
    elif output_type == "pbox":
        return pbox_from_ecdf_bundle(b_l, b_r)
    elif output_type == "un":
        p = pbox_from_ecdf_bundle(b_l, b_r)
        return UncertainNumber.from_pbox(p)


def logical_bounding(a):
    """Sudret p16. eq(2.21)"""
    a = np.where(a < 0, 0, a)
    a = np.where(a < 1, a, 1)
    return a


def d_alpha(n, alpha):
    """compute the Smirnov critical value for a given sample size and significance level

    note:
        Tretiak p12. eq(8): alpha = (1-c) / 2 where c is the confidence level

    args:
        - n (int): sample size;
        - alpha (float): significance level;
    """

    A = {0.1: 0.00256, 0.05: 0.05256, 0.025: 0.11282}
    return (
        np.sqrt(np.log(1 / alpha) / (2 * n))
        - 0.16693 * (1 / n)
        - A.get(alpha, 1000) * (n ** (-3 / 2))
    )


# * ---------top level func for known statistical properties------*#


@exposeUN
def known_properties(
    maximum=None,
    mean=None,
    median=None,
    minimum=None,
    mode=None,
    percentiles=None,
    std=None,
    var=None,
    family=None,
    **kwargs,
) -> UncertainNumber:
    """Construct a uncertain number given known statistical properties served as constraints.

    args:
        maximum (number): maximum value of the variable
        mean (number): mean value of the variable
        median (number): median value of the variable
        minimum (number): minimum value of the variable
        mode (number): mode value of the variable
        percentiles (dict): dictionary of percentiles and their values, e.g. {0: 0, 0.1: 1, 0.5: 2, 0.9: pun.I(3,4), 1:5}
        std (number): standard deviation of the variable
        var (number): variance of the variable
        family (str): name of the distribution family, e.g. 'normal', 'lognormal', 'uniform', 'triangular', etc.

    returns:
        uncertain number

    note:
        It's also possible to directly call a function given the known information, such as ``pun.mean_std(mean=1, std=0.5)``.

    example:
        >>> from pyuncertainnumber.pba import known_properties
        >>> known_properties(
        ...     maximum = 2,
        ...     mean = 1,
        ...     var = 0.25,
        ...     minimum=0,
        ...     )
    """

    from ..characterisation.stats import parse_moments

    args = {
        "maximum": maximum,
        "mean": mean,
        "median": median,
        "minimum": minimum,
        "mode": mode,
        "percentiles": percentiles,
        "std": std,
        "var": var,
        "family": family,
    }
    shape_control = ["percentiles", "symmetric"]
    present_keys = tuple(
        sorted(k for k, v in args.items() if v is not None if k not in shape_control)
    )

    # template:
    # ('a', 'b'): handle_ab,

    routes = {
        ("percentiles"): from_percentiles,
        ("maximum", "minimum"): min_max,
        ("mean", "minimum"): min_mean,
        ("maximum", "mean"): max_mean,
        ("mean", "std"): mean_std,
        ("mean", "var"): mean_var,
        ("maximum", "mean", "minimum"): min_max_mean,
        ("maximum", "minimum", "mode"): min_max_mode,
        ("maximum", "median", "minimum"): min_max_median,
        ("maximum", "mean", "minimum", "std"): min_max_mean_std,
        ("maximum", "mean", "minimum", "var"): min_max_mean_var,
        ("family"): not_enough_info,
        ("family", "mean"): parse_moments,
        ("family", "mean", "std"): parse_moments,
        ("family", "mean", "var"): parse_moments,
        ("family", "mean", "std", "var"): parse_moments,
        (
            "family",
            "maximum",
            "minimum",
        ): min_max,  # TODO: to implement trucate_parse_moments
        (
            "family",
            "maximum",
            "mean",
            "minimum",
        ): parse_moments,
        ("family", "maximum", "mean", "minimum", "std"): truncate_parse_moments,
        ("family", "maximum", "mean", "minimum", "var"): truncate_parse_moments,
        ("family", "maximum", "mean", "minimum", "std", "var"): truncate_parse_moments,
    }

    handler1 = routes.get(present_keys, handle_default)
    base_pbox = handler1(**{k: args[k] for k in present_keys})

    # second-level shape control to see if percentiles or some other constraints are present
    further_shape_controls = [
        k for k, v in args.items() if v is not None if k in shape_control
    ]

    if not further_shape_controls:
        return base_pbox
    else:
        for c_keys in further_shape_controls:
            c_handler = routes.get(c_keys, handle_default)
            c_pbox = c_handler(args[c_keys])
            if not isinstance(base_pbox, Pbox):
                return c_pbox
            imp_pbox = base_pbox.imp(c_pbox)
        return imp_pbox


known_constraints = known_properties


def handle_default(**kwargs):
    raise Exception(f"Combination not supported. Received: {kwargs}")


def not_enough_info(**kwargs):
    raise Exception(f"Not enough information provided. Received: {kwargs}")


def truncate_parse_moments(**kwargs):
    from ..characterisation.stats import parse_moments
    from .operation import convert

    if "maximum" in kwargs and "minimum" in kwargs:
        base_pbox = convert(parse_moments(**kwargs))
        box = min_max(**{k: kwargs[k] for k in ["minimum", "maximum"] if k in kwargs})
        if base_pbox.hi <= box.hi and base_pbox.lo >= box.lo:
            return base_pbox.imp(box)
        else:
            raise Exception("No intersection found")

    else:
        return parse_moments(**kwargs)


# * --------------------- supporting functions---------------------*#


def min_max(minimum: Number, maximum: Number) -> UncertainNumber | Pbox:
    """Equivalent to an interval object constructed as a nonparametric Pbox.

    args:
        minimum : Left end of box
        maximum : Right end of box

    returns:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> from pyuncertainnumber.pba import min_max
        >>> min_max(0, 2)  # return a UncertainNumber
        >>> min_max(0, 2, return_construct=True)  # return a Pbox
    """

    return Staircase(
        left=np.repeat(minimum, Params.steps),
        right=np.repeat(maximum, Params.steps),
        mean=I(minimum, maximum),
        var=I(0, (maximum - minimum) * (maximum - minimum) / 4),
    )


def min_mean(minimum, mean, steps=Params.steps) -> UncertainNumber | Pbox:
    """Nonparametric pbox construction based on constraint of minimum and mean

    args:
        minimum (number): minimum value of the variable
        mean (number): mean value of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> from pyuncertainnumber.pba import min_mean
        >>> min_mean(0, 1)  # return a UncertainNumber
        >>> min_mean(0, 1, return_construct=True)  # return a Pbox
    """
    jjj = np.array([j / steps for j in range(1, steps - 1)] + [1 - 1 / steps])
    right = [((mean - minimum) / (1 - j) + minimum) for j in jjj]

    return Staircase(
        left=np.repeat(minimum, len(right)),
        right=right,
        mean=I(mean, mean),
    )


def max_mean(
    maximum: Number,
    mean: Number,
    steps=Params.steps,
) -> UncertainNumber | Pbox:
    # TODO no __neg__
    """Nonparametric pbox construction based on constraint of maximum and mean

    args:
        maximum (number): maximum value of the variable
        mean (number): mean value of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> max_mean(2, 1)  # return a UncertainNumber
    """
    return min_mean(-maximum, -mean).__neg__()


def mean_std(mean: Number, std: Number, steps=Params.steps) -> UncertainNumber | Pbox:
    """Nonparametric pbox construction based on constraint of mean and std

    args:
        mean (number): mean value of the variable
        std (number): std value of the variable

    return:
        UncertainNumber or Pbox


    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.


    example:
        >>> mean_std(1, 0.5)
    """
    iii = [1 / steps] + [i / steps for i in range(1, steps - 1)]
    jjj = [j / steps for j in range(1, steps - 1)] + [1 - 1 / steps]

    left = [mean - std * np.sqrt(1 / i - 1) for i in iii]
    right = [mean + std * np.sqrt(j / (1 - j)) for j in jjj]
    return Staircase(left=left, right=right, mean=I(mean, mean), var=I(std**2, std**2))


def mean_var(
    mean: Number,
    var: Number,
) -> UncertainNumber | Pbox:
    """Nonparametric pbox construction based on constraint of mean and var

    args:
        mean (number): mean value of the variable
        vasr (number): var value of the variable

    return:
        UncertainNumber or Pbox


    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> mean_var(1, 0.25)  # return a UncertainNumber
    """
    return mean_std(mean, np.sqrt(var))


def min_max_mean(
    minimum: Number,
    maximum: Number,
    mean: Number,
    steps: int = Params.steps,
) -> UncertainNumber | Pbox:
    # TODO var is missing
    """Generates a distribution-free p-box based upon the minimum, maximum and mean of the variable

    args:
        minimum (float): minimum value of the variable
        maximum (float): maximum value of the variable
        mean (float): mean value of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.


    example:
        >>> min_max_mean(0, 2, 1)
    """
    mid = (maximum - mean) / (maximum - minimum)
    ii = [i / steps for i in range(steps)]
    left = [minimum if i <= mid else ((mean - maximum) / i + maximum) for i in ii]
    jj = [j / steps for j in range(1, steps + 1)]
    right = [maximum if mid <= j else (mean - minimum * j) / (1 - j) for j in jj]
    # print(len(left))
    return Staircase(
        left=np.array(left), right=np.array(right), mean=I(mean, mean), steps=steps
    )


# TODO: to verify if this is correct
def pos_mean_std(
    mean: Number,
    std: Number,
    steps=Params.steps,
) -> Pbox:
    """Generates a positive distribution-free p-box based upon the mean and standard deviation of the variable

    args:
        mean : mean of the variable
        std : standard deviation of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    """
    iii = [1 / steps] + [i / steps for i in range(1, steps - 1)]
    jjj = [j / steps for j in range(1, steps - 1)] + [1 - 1 / steps]

    left = [max((0, mean - std * np.sqrt(1 / i - 1))) for i in iii]
    right = [min((mean / (1 - j), mean + std * np.sqrt(j / (1 - j)))) for j in jjj]

    return Staircase(
        left=left,
        right=right,
        steps=steps,
        mean=I(mean, mean),
        var=I(std**2, std**2),
    )


def min_max_mode(
    minimum: Number,
    maximum: Number,
    mode: Number,
    steps: int = Params.steps,
) -> UncertainNumber | Pbox:
    """Nonparametric pbox construction based on constraint of mean and var

    args:
        minimum: minimum value of the variable
        maximum: maximum value of the variable
        mode (number): mode value of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> min_max_mode(0, 2, 1)  # return a UncertainNumber
    """
    if minimum == maximum:
        return min_max(minimum, maximum)

    ii = np.array([i / steps for i in range(steps)])
    jj = np.array([j / steps for j in range(1, steps + 1)])

    l = ii * (mode - minimum) + minimum
    r = jj * (maximum - mode) + mode
    mean_l = (minimum + mode) / 2
    mean_r = (mode + maximum) / 2
    var_l = 0
    var_r = (maximum - minimum) * (maximum - minimum) / 12

    return Staircase(left=l, right=r, mean=I(mean_l, mean_r), var=I(var_l, var_r))


def min_max_median(
    minimum: Number,
    maximum: Number,
    median: Number,
    steps: int = Params.steps,
) -> UncertainNumber | Pbox:
    """Generates a distribution-free p-box based upon the minimum, maximum and median of the variable

    args:
        minimum : minimum value of the variable
        maximum : maximum value of the variable
        median : median value of the variable

    return:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> min_max_median(0, 2, 1)  # return a UncertainNumber

    """
    if minimum == maximum:
        return min_max(minimum, maximum)

    p_minmax = I(minimum, maximum).to_pbox()
    r = p_minmax.alpha_cut(0.5)
    l = p_minmax.left.copy()

    half_mark = steps // 2

    l_quantile = np.where(p_minmax.p_values < 0.5, minimum, median)
    r_quantile = np.where(p_minmax.p_values >= 0.5, maximum, median)
    return Staircase(
        left=l_quantile,
        right=r_quantile,
        mean=I((minimum + median) / 2, (median + maximum) / 2),
        var=I(0, (maximum - minimum) * (maximum - minimum) / 4),
        steps=steps,
    )


def min_max_mean_std(
    minimum: Number,
    maximum: Number,
    mean: Number,
    std: Number,
    **kwargs,
) -> UncertainNumber | Pbox:
    """Generates a distribution-free p-box based upon the minimum, maximum, mean and standard deviation of the variable

    args:
        maximum (number): maximum value of the variable
        minimum (number): minimum value of the variable
        std (number): standard deviation of the variable
        var (number): variance of the variable

    return:
        UncertainNumber or Pbox


    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.


    example:
        >>> min_max_mean_std(0, 2, 1, 0.5)  # return a UncertainNumber

    .. seealso::

        :func:`min_max_mean_var`

    """
    if minimum == maximum:
        return min_max(minimum, maximum)

    def _left(x):

        if isinstance(x, (int, float, np.number)):
            return x
        if x.__class__.__name__ == "Interval":
            return x.left
        if x.__class__.__name__ == "Pbox":
            return min(x.left)
        else:
            raise Exception("wrong type encountered")

    def _right(x):
        if isinstance(x, (int, float, np.number)):
            return x
        if x.__class__.__name__ == "Interval":
            return x.right
        if x.__class__.__name__ == "Pbox":
            return max(x.right)

    def _imp(a, b):
        return I(max(_left(a), _left(b)), min(_right(a), _right(b)))

    def _env(a, b):
        return I(min(_left(a), _left(b)), max(_right(a), _right(b)))

    def _constrain(a, b, msg):
        if (_right(a) < _left(b)) or (_right(b) < _left(a)):
            print("Math Problem: impossible constraint", msg)
        return _imp(a, b)

    zero = 0.0
    one = 1.0
    ran = maximum - minimum
    m = _constrain(mean, I(minimum, maximum), "(mean)")
    s = _constrain(
        std,
        _env(
            I(0.0),
            (abs(ran * ran / 4.0 - (maximum - mean - ran / 2.0) ** 2)) ** 0.5,
        ),
        " (dispersion)",
    )
    ml = (m.left - minimum) / ran
    sl = s.left / ran
    mr = (m.right - minimum) / ran
    sr = s.right / ran
    z = I(minimum, maximum).to_pbox()
    n = len(z.left)
    L = [0.0] * n
    R = [1.0] * n
    for i in range(n):
        p = i / n
        if p <= zero:
            x2 = zero
        else:
            x2 = ml - sr * (one / p - one) ** 0.5
        if ml + p <= one:
            x3 = zero
        else:
            x5 = p * p + sl * sl - p
            if x5 >= zero:
                x4 = one - p + x5**0.5
                if x4 < ml:
                    x4 = ml
            else:
                x4 = ml
            x3 = (p + sl * sl + x4 * x4 - one) / (x4 + p - one)
        if (p <= zero) or (p <= (one - ml)):
            x6 = zero
        else:
            x6 = (ml - one) / p + one
        L[i] = max(max(max(x2, x3), x6), zero) * ran + minimum

        p = (i + 1) / n
        if p >= one:
            x2 = one
        else:
            x2 = mr + sr * (one / (one / p - one)) ** 0.5
        if mr + p >= one:
            x3 = one
        else:
            x5 = p * p + sl * sl - p
            if x5 >= zero:
                x4 = one - p - x5**0.5
                if x4 > mr:
                    x4 = mr
            else:
                x4 = mr
            x3 = (p + sl * sl + x4 * x4 - one) / (x4 + p - one) - one

        if ((one - mr) <= p) or (one <= p):
            x6 = one
        else:
            x6 = mr / (one - p)
        R[i] = min(min(min(x2, x3), x6), one) * ran + minimum

    v = s**2
    return Staircase(
        left=np.array(L),
        right=np.array(R),
        mean=I(_left(m), _right(m)),
        var=I(_left(v), _right(v)),
        **kwargs,
    )


def min_max_mean_var(
    minimum: Number,
    maximum: Number,
    mean: Number,
    var: Number,
    **kwargs,
) -> UncertainNumber | Pbox:
    """Generates a distribution-free p-box based upon the minimum, maximum, mean and standard deviation of the variable

    args:
        minimum (number): minimum value of the variable
        maximum (number): maximum value of the variable
        mean (number): mean value of the variable
        var (number): variance of the variable


    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    example:
        >>> min_max_mean_var(0, 2, 1, 0.25)  # return a UncertainNumber

    .. admonition:: Implementation

        Equivalent to ``min_max_mean_std(minimum,maximum,mean,np.sqrt(var))``

    .. seealso::

        :func:`min_max_mean_std`

    """
    return min_max_mean_std(minimum, maximum, mean, np.sqrt(var), **kwargs)


def from_percentiles(
    percentiles: dict, steps: int = Params.steps
) -> UncertainNumber | Pbox:
    """yields a distribution-free p-box based on specified percentiles of the variable

    args:
        percentiles : dictionary of percentiles and their values (e.g. {0: 0, 0.1: 1, 0.5: 2, 0.9: I(3,4), 1:5})
        steps : number of steps to use in the p-box

    note:
        The percentiles dictionary is of the form {percentile: value}. Where value can either be a number or an I. If value is a number, the percentile is assumed to be a point percentile. If value is an I, the percentile is assumed to be an interval percentile.
        If no keys for 0 and 1 are given, ``-np.inf`` and ``np.inf`` are used respectively. This will result in a p-box that is not bounded and raise a warning.
        If the percentiles are not increasing, the percentiles will be intersected. This may not be desired behaviour.
        ValueError: If any of the percentiles are not between 0 and 1.

    returns:
        UncertainNumber or Pbox

    tip:
        Two types of return values are possible:

        - by default, a `UncertainNumber` is returned;

        - For low-level controls, if `return_construct=True` is specified, a `Pbox` is returned.

    Example:
        >>> pba.from_percentiles(
        >>>     {0: 0,
        >>>     0.25: 0.5,
        >>>     0.5: pba.I(1,2),
        >>>     0.75: pba.I(1.5,2.5),
        >>>     1: 3})
        >>>     .display()
    """
    # check if 0 and 1 are in the dictionary
    if 0 not in percentiles.keys():
        percentiles[0] = -np.inf
        warn("No value given for 0 percentile. Using -np.inf")
    if 1 not in percentiles.keys():
        percentiles[1] = np.inf
        warn("No value given for 1 percentile. Using np.inf")

    # sort the dictionary by percentile
    percentiles = dict(sorted(percentiles.items()))

    from .intervals.intervalOperators import wc_scalar_interval

    # transform values to intervals
    for k, v in percentiles.items():
        # if not isinstance(v, I):
        percentiles[k] = wc_scalar_interval(v)

    if any([p < 0 or p > 1 for p in percentiles.keys()]):
        raise ValueError("Percentiles must be between 0 and 1")

    left = []
    right = []
    for i in np.linspace(0, 1, steps):
        smallest_key = min(key for key in percentiles.keys() if key >= i)
        largest_key = max(key for key in percentiles.keys() if key <= i)
        left.append(percentiles[largest_key].left)
        right.append(percentiles[smallest_key].right)

    try:
        # return Pbox(left, right, steps=steps, interpolation="outer")  # backup
        return Staircase(left=left, right=right, steps=steps)
    except NotIncreasingError:
        warn("Percentiles are not increasing. Will take intersection of percentiles.")

        left = []
        right = []
        p = list(percentiles.keys())

        def sometimes(condition):
            """dummy"""
            pass

        for i, j, k in zip(p, p[1:], p[2:]):
            if sometimes(percentiles[j] < percentiles[i]):
                percentiles[j] = I(percentiles[i].right, percentiles[j].right)
            if sometimes(percentiles[j] > percentiles[k]):
                percentiles[j] = I(percentiles[j].left, percentiles[k].left)

        left = []
        right = []
        for i in np.linspace(0, 1, steps):
            smallest_key = min(key for key in percentiles.keys() if key >= i)
            left.append(percentiles[smallest_key].left)
            right.append(percentiles[smallest_key].right)

        # return Pbox(left, right, steps=steps, interpolation="outer")  # backup
        return Staircase(left=left, right=right, steps=steps)
    except:
        raise Exception("Unable to generate p-box")



# # existing exports...
# from .utils import (
#     CalibUtils,
#     make_spd,
#     get_gaussian_emp_stats,
#     compute_ecdf,
#     plot_ecdf_comparison,
#     plot_top_by_quantile_scatter,
#     kde_1d,
# )

# __all__ = [
#     # ...existing...
#     "CalibUtils",
#     "make_spd",
#     "get_gaussian_emp_stats",
#     "compute_ecdf",
#     "plot_ecdf_comparison",
#     "plot_top_by_quantile_scatter",
#     "kde_1d",
# ]

from .tmcmc import TMCMC, Stage
from .calibration import MCMCCalibrator
from .knn import KNNCalibrator
from .epistemic_filter import EpistemicFilter



from __future__ import annotations
from typing import TYPE_CHECKING
import functools

from pyuncertainnumber.pba.operation import convert
from .un_fields import *
from .utils import *
from .config import Config
from pathlib import Path
from ..nlp.language_parsing import hedge_interpret
from .check import DistributionSpecification
from ..pba.pbox_parametric import named_pbox
from ..pba.intervals.intervalOperators import parse_bounds, wc_scalar_interval_feature
from ..pba.intervals.number import Interval
from numbers import Number
from ..pba.distributions import Distribution as pbaDistribution
import operator
from pint import Quantity


""" Uncertain Number class """


if TYPE_CHECKING:
    from ..pba.intervals.number import Interval
    from ..pba.distributions import Distribution as pbaDistribution


__all__ = [
    "UncertainNumber",
    "I",
    "D",
    "DSS",
    "norm",
    "gaussian",
    "gamma",
    "normal",
    "alpha",
    "anglit",
    "argus",
    "arcsine",
    "beta",
    "betaprime",
    "bradford",
    "burr",
    "burr12",
    "cauchy",
    "chi",
    "chi2",
    "cosine",
    "uniform",
    "lognormal",
]


class UncertainNumber:
    """Uncertain Number class

    args:
        - intervals (Interval): the interval specification for the UN object;

        - distribution_parameters: a list of the distribution family and its parameters; e.g. ['norm', [0, 1]];

        - pbox_initialisation: a list of the distribution family and its parameters; e.g. ['norm', ([0,1], [3,4])];


    example:

        Uncertain numbers can be constructed in multiple ways. For example, a canonical way allows users to fill in as many fields as possible:

        >>> from pyuncertainnumber import UncertainNumber
        >>> UncertainNumber(name="velocity", symbol="v", unit="m/s", intervals=[1, 2])
        >>> UncertainNumber(name="velocity", symbol="v", unit="m/s", distribution_parameters=['normal', (10, 2)])
        >>> UncertainNumber(name="velocity", symbol="v", unit="m/s", pbox_parameters=['normal', ([8, 12], [0.5, 1.5])])
        >>> UncertainNumber(name="velocity", symbol="v", unit="m/s", essence='dempster_shafer', intervals=[[1,5], [3,6]], masses=[0.5, 0.5])


        Alternatively, users can use shortcuts to quickly create UN objects and get on with calculations:

        >>> import pyuncertainnumber as pun
        >>> pun.I([1, 2])
        >>> pun.D('gaussian', (10, 2))
        >>> pun.normal([8, 12], [0.5, 1.5])
        >>> pun.DSS(intervals=[[1,5], [3,6]], masses=[0.5, 0.5])
    """

    Q_ = Quantity
    instances = []

    def __init__(
        self,
        name=None,
        symbol=None,
        unit=None,
        uncertainty_type=None,
        essence=None,
        masses=None,
        intervals=None,
        distribution_parameters=None,
        pbox_parameters=None,
        hedge=None,
        _construct=None,
        nominal_value=1.0,
        p_flag=True,
        _skip_construct_init=False,
        measurand=None,
        nature=None,
        provenence=None,
        justification=None,
        structure=None,
        security=None,
        ensemble=None,
        variability=None,
        dependence=None,
        uncertainty=None,
        physical_quantity=None,
        _samples=None,
        **kwargs,
    ):
        self.name = name
        self.symbol = symbol
        self.uncertainty_type = uncertainty_type
        self.essence = essence
        self.masses = masses
        self.intervals = intervals
        self.distribution_parameters = distribution_parameters
        self.pbox_parameters = pbox_parameters
        self.hedge = hedge
        self._construct = _construct
        self.nominal_value = nominal_value
        self.p_flag = p_flag
        self._skip_construct_init = _skip_construct_init
        self.measurand = measurand
        self.nature = nature
        self.provenence = provenence
        self.justification = justification
        self.structure = structure
        self.security = security
        self.ensemble = ensemble
        self.variability = variability
        self.dependence = dependence
        self.uncertainty = uncertainty
        self._physical_quantity = physical_quantity
        self._samples = _samples
        self.__init_check()
        self.__init_construct()
        self.nominal_value = self._construct.nominal_value
        self.unit = unit

    # *  ---------------------more on initialisation---------------------*#

    def __init_check(self):
        UncertainNumber.instances.append(self)

        if not self.essence:
            check_initialisation_list = [
                self.intervals,
                self.distribution_parameters,
                self.pbox_parameters,
            ]
            if any(v is not None for v in check_initialisation_list):
                raise ValueError(
                    "The 'essence' of the Uncertain Number is not specified"
                )
            if (self._construct is None) | (not self._construct):
                print("a vacuous interval is created")
                self.essence = "interval"
                self.intervals = [-np.inf, np.inf]

    def __init_construct(self):
        """the de facto parameterisation/instantiation procedure for the core constructs of the UN class

        caveat:
            user needs to by themselves figure out the correct
            shape of the 'distribution_parameters', such as ['uniform', [1,2]]
        """

        # * ------------------------ create the underlying construct

        if not self._skip_construct_init:
            match self.essence:
                case "interval":
                    self._construct = parse_bounds(self.intervals)
                case "distribution" | "pbox":
                    if self.pbox_parameters is not None:
                        par = Parameterisation(
                            self.pbox_parameters, essence=self.essence
                        )
                    else:
                        par = Parameterisation(
                            self.distribution_parameters, essence=self.essence
                        )
                    self._construct = par.yield_construct()
                case "dempster_shafer":
                    from ..pba.dss import DempsterShafer

                    self._construct = DempsterShafer(
                        intervals=parse_bounds(self.intervals), masses=self.masses
                    )

    def parameterised_pbox_specification(self):
        if self.p_flag:
            self._construct = self.match_pbox(
                self.distribution_parameters[0],
                self.distribution_parameters[1],
            )
            self.nominal_value = self._construct.mean().midpoint()

    def _update_physical_quantity(self):
        self._physical_quantity = self.Q_(self.nominal_value, self.unit)

    @staticmethod
    def match_pbox(keyword, parameters):
        """match the distribution keyword from the initialisation to create the underlying distribution object

        args:
            - keyword: (str) the distribution keyword
            - parameters: (list) the parameters of the distribution
        """
        obj = named_pbox.get(
            keyword, "You're lucky as the distribution is not supported"
        )
        if isinstance(obj, str):
            print(obj)  # print the error message
        return obj(*parameters)

    def init_check(self):
        """check if the UN initialisation specification is correct

        note:
            a lot of things to double check. keep an growing list:
            1. unit
            2. hedge: user cannot speficy both 'hedge' and 'intervals'. 'intervals' takes precedence.

        """
        pass

    # * ---------------------object representation---------------------* #

    def __str__(self):
        """string representation of the UncertainNumber

        note:
            the same as __reor__ for now until a better idea is proposed
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """Concise __repr__"""

        field_values = get_concise_repr(self.__dict__)
        field_str = ", ".join(
            f"{k}={float(v)!r}" if isinstance(v, np.floating) else f"{k}={v!r}"
            for k, v in field_values.items()
        )

        # fancy string formatting of unit
        u_str = f", physical_quantity={self._physical_quantity:~P}"
        if not self._physical_quantity.dimensionless:
            field_str += u_str

        return f"{self.__class__.__name__}({field_str})"

    def describe(self, style="verbose"):
        """print out a verbose description of the uncertain number"""

        match style:
            case "verbose":
                match self.essence:
                    case "interval":
                        return f"This is an {self.essence}-type Uncertain Number whose min value is {self._construct.left:.2f} and max value is {self._construct.right:.2f}. An interval is a range of values that are possible for the measurand whose value is unknown, which typically represents the epistemic uncertainty. The interval is defined by the minimum and maximum values (i.e. lower bound and upper bound) that the measurand could take on."
                    case "distribution":
                        return f"This is a {self.essence}-type Uncertain Number that follows a {self.distribution_parameters[0]} distribution with parameters {self.distribution_parameters[1]}. Probability distributios are typically empolyed to model aleatoric uncertainty, which represents inherent randomness. The distribution is defined by the probability density function (pdf) or cumulative distribution function (cdf)."
                    case "pbox":
                        try:
                            return f"This is a {self.essence}-type Uncertain Number that follows a {self.pbox_parameters[0]} distribution with parameters {self.pbox_parameters[1]}"
                        except:
                            return f"This is a {self.essence}-type Uncertain Number that follows a {self.distribution_parameters[0]} distribution with parameters {self.distribution_parameters[1]}"

            case "one-number":
                return f"This is an {self.essence}-type Uncertain Number whose naked value is {self.nominal_value:.2f}"
            case "concise":
                return self.__repr__()
            case "range":
                match self.essence:
                    case "interval":
                        return f"This is an {self.essence}-type Uncertain Number whose min value is {self._construct.left:.2f} and max value is {self._construct.right:.2f}."
                    case "distribution":
                        return f"This is an {self.essence}-type Uncertain Number with 'some' range of {self._construct._range_list[0]:.2f} and {self._construct._range_list[1]:.2f}."
                    case "pbox":
                        return f"This is an {self.essence}-type Uncertain Number with 'some' range of {self._construct.left:.2f} and {self._construct.right:.2f}."
            case "five-number":
                match self.essence:
                    case "interval":
                        return f"This is an {self.essence}-type Uncertain Number that does not support this description."
                    case "distribution":
                        print(
                            f"This is an {self.essence}-type Uncertain Number whose statistical description is shown below:\n"
                            f"- family: {self.distribution_parameters[0]}\n"
                            f"- min: {self._construct._range_list[0]:.2f}\n"
                            f"- Q1: something\n"
                            f"- mean: {self._construct.mean_left}\n"
                            f"- Q3: something\n"
                            f"- variance: something"
                        )
            case "risk calc":
                match self.essence:
                    case "interval":
                        return "Will show a plot of the interval"
                    case "distribution":
                        print(
                            f"This is an {self.essence}-type Uncertain Number of family '{self.distribution_parameters[0]}' parameterised by {self.distribution_parameters[1]}"
                        )
                        self._construct.quick_plot()

    # * ---------------------some methods---------------------* #

    def ci(self):
        """get 95% range confidence interval"""
        match self.essence:
            case "interval":
                return [self._construct.left, self._construct.right]
            case "distribution":
                which_dist = self.distribution_parameters[0]
                if which_dist == "norm":
                    rv = norm(*self.distribution_parameters[1])
                    return [rv.ppf(0.025), rv.ppf(0.975)]
            case "pbox":
                return "unfinshed"

    def plot(self, **kwargs):
        """quick plot of the uncertain number object"""

        return self._construct.plot(**kwargs)

    def display(self, **kwargs):
        """quick plot of the uncertain number object"""

        return self._construct.display(**kwargs)

    # * ---------------------properties --------------------- *#
    @property
    def construct(self):
        return self._construct

    @property
    def construct_type(self):
        type(self._construct)

    @property
    def unit(self):
        """get the physical quantity of the uncertain number"""
        return self._units

    @unit.setter
    def unit(self, value):
        """set the physical quantity of the uncertain number"""
        self._units = value
        self._update_physical_quantity()

    @property
    def physical_quantity(self):
        """get the physical quantity of the uncertain number"""
        return self._physical_quantity

    @physical_quantity.setter
    def physical_quantity(self, value):
        """set the physical quantity of the uncertain number"""
        self._physical_quantity = value

    # * --------------------- constructors--------------------- *#

    @classmethod
    def from_hedge(cls, hedged_language):
        """create an Uncertain Number from hedged language"""
        an_obj = hedge_interpret(hedged_language)
        essence = "interval"  # TODO: choose between interval, pbox
        left, right = an_obj.left, an_obj.right
        return cls(essence=essence, intervals=[left, right])

    @classmethod
    def fromConstruct(cls, construct):
        """create an Uncertain Number from a construct object"""
        from ..pba.pbox_abc import Pbox
        from ..pba.dss import DempsterShafer
        from ..pba.distributions import Distribution as pbaDistribution

        if isinstance(construct, Pbox):
            return cls.from_pbox(construct)
        if isinstance(construct, Interval):
            return cls.from_Interval(construct)
        if isinstance(construct, DempsterShafer):
            return cls.from_ds(construct)
        if isinstance(construct, pbaDistribution):
            return cls.fromDistribution(construct)
        if isinstance(construct, cls):
            return construct
        if isinstance(
            construct, np.ndarray
        ):  # a fail-safe exit, which may be coerced into a Distribution UN later on
            return construct
        else:
            raise ValueError("The construct object is not recognised")

    @classmethod
    def fromDistribution(cls, D, **kwargs):
        # dist_family: str, dist_params,
        """create an Uncertain Number from a Distribution object.

        args:
            - D (Distribution): a Distribution object
            dist_family (str): the distribution family
            dist_params (list, tuple or string): the distribution parameters
        """
        # distSpec = DistributionSpecification(D.dist_family, D.dist_params)

        if D.empirical_data is None:
            return cls(
                essence="distribution",
                distribution_parameters=[D.dist_family, D.dist_params],
                **kwargs,
            )
        else:
            return cls(
                essence="distribution",
                distribution_parameters=None,
                _samples=D.sample_data,
            )

    @classmethod
    def from_Interval(cls, u):
        return cls(essence="interval", intervals=u)

    @classmethod
    def from_pbox(cls, p):
        """genenal from  pbox"""
        # passPboxParameters()
        return cls(
            essence="pbox", p_flag=False, _construct=p, _skip_construct_init=True
        )

    @classmethod
    def from_ds(cls, ds):
        return cls.from_pbox(ds.to_pbox())

    @classmethod
    def from_sps(cls, sps_dist):
        """create an UN object from a parametric scipy.stats dist object
        #! it seems that a function will suffice
        args:
            - sps_dist: scipy.stats dist object

        note:
            - sps_dist --> UN.Distribution object
        """
        pass

    # * ---------------------arithmetic operations---------------------#

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented
        if len(inputs) != 1 or inputs[0] is not self:
            return NotImplemented
        if "out" in kwargs and kwargs["out"] is not None:
            return NotImplemented

        if ufunc is np.sin:
            return self.sin()
        if ufunc is np.cos:
            return self.cos()
        if ufunc is np.tan:
            return self.tan()
        if ufunc is np.tanh:
            return self.tanh()
        if ufunc is np.exp:
            return self.exp()
        if ufunc is np.sqrt:
            return self.sqrt()
        if ufunc is np.log:
            return self.log()
        if ufunc is np.reciprocal:
            return self.reciprocal()

        return NotImplemented

    # * ---------------------unary operations---------------------#

    def _apply(self, method):
        from ..pba.operation import convert

        target = convert(self.construct)
        return UncertainNumber.fromConstruct(getattr(target, method)())

    def sqrt(self):
        return self._apply("sqrt")

    def exp(self):
        return self._apply("exp")

    def tanh(self):
        return self._apply("tanh")

    def tan(self):
        return self._apply("tan")

    def log(self):
        return self._apply("log")

    def sin(self):
        return self._apply("sin")

    def cos(self):
        return self._apply("cos")

    def reciprocal(self):
        return self._apply("reciprocal")

    # * ---------------------binary operations---------------------#

    def bin_ops(self, other, ops):
        from ..pba.pbox_abc import convert_pbox

        # new_cons = ops(self._construct, other._construct)
        if is_un(other) == 0:
            new_cons = ops(self._construct, other)
        elif is_un(other) == 1:
            a = convert_pbox(self._construct)
            b = convert_pbox(other._construct)
            new_cons = ops(a, b)
        elif is_un(other) == 2:
            raise ValueError(
                "construct object entered. propagation can be done but units won't be passed down"
            )

        new_un = UncertainNumber.fromConstruct(new_cons)
        return pass_down_units(self, other, ops, new_un)

    def __add__(self, other):
        """add two uncertain numbers"""
        return self.bin_ops(other, operator.add)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.bin_ops(other, operator.sub)

    def __mul__(self, other):
        """multiply two uncertain numbers"""
        return self.bin_ops(other, operator.mul)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        """divide two uncertain numbers"""
        return self.bin_ops(other, operator.truediv)

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __pow__(self, other):
        """power of two uncertain numbers"""
        return self.bin_ops(other, operator.pow)

    def __rpow__(self, other):
        """power of two uncertain numbers"""
        from ..pba.operation import convert

        return other ** convert(self.construct)

    # * ---------------------w/ dependency ---------------------#

    def add(self, other, dependency="f"):
        from ..pba.operation import convert

        a, b = convert(self.construct), convert(other.construct)
        return a.add(b, dependency=dependency)

    def sub(self, other, dependency="f"):
        from ..pba.operation import convert

        a, b = convert(self.construct), convert(other.construct)
        return a.sub(b, dependency=dependency)

    def mul(self, other, dependency="f"):
        from ..pba.operation import convert

        a, b = convert(self.construct), convert(other.construct)
        return a.mul(b, dependency=dependency)

    def div(self, other, dependency="f"):
        from ..pba.operation import convert

        a, b = convert(self.construct), convert(other.construct)
        return a.div(b, dependency=dependency)

    def pow(self, other, dependency="f"):
        from ..pba.operation import convert

        a, b = convert(self.construct), convert(other.construct)
        return a.pow(b, dependency=dependency)

    # * ---------------------serialisation functions---------------------*#

    def JSON_dump(self, filename="UN_data.json"):
        """the JSON serialisation of the UN object into the filesystem"""

        filepath = Path(Config.result_path) / filename
        with open(filepath, "w") as fp:
            json.dump(self, fp, cls=UNEncoder, indent=4)

    # * ---------------------other functions---------------------*#
    def to_pbox(self):
        """convert the UN object to a pbox object"""

        from ..pba.operation import convert

        pbox = convert(self._construct)
        return pbox


# * ---------------------shortcuts --------------------- *#
def makeUNPbox(func):
    """Constructor decorator to create a UN from a parametric pbox"""
    from ..pba.pbox_parametric import _bound_pcdf

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        family_str = func(*args, **kwargs)
        p = _bound_pcdf(family_str, *args)
        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


def constructUN(func):
    """from a construct to create a UN"""

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        p = func(*args, **kwargs)
        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


def I(*args: str | list[Number] | Interval) -> UncertainNumber:
    """a shortcut to construct the interval-type UN object"""
    return UncertainNumber.fromConstruct(wc_scalar_interval_feature(*args))


def D(*args, **kwargs) -> UncertainNumber:
    """a shortcut to construct the distribution-type UN object"""
    from ..pba.distributions import Distribution as pbaDistribution

    dist = pbaDistribution(*args, **kwargs)
    return UncertainNumber.fromDistribution(dist)


def DSS(*args, **kwargs) -> UncertainNumber:
    """a shortcut for the Dempster-Shafer-type UN object"""
    from ..pba.dss import DempsterShafer

    ds = DempsterShafer(*args, **kwargs)
    return UncertainNumber.from_ds(ds)


# * ---------------------parse inputs for UN only  --------------------- *#


def match_pbox(keyword, parameters):
    """match the distribution keyword from the initialisation to create the underlying distribution object

    args:
        - keyword: (str) the distribution keyword
        - parameters: (list) the parameters of the distribution
    """
    obj = named_pbox.get(keyword, "You're lucky as the distribution is not supported")
    if isinstance(obj, str):
        print(obj)  # print the error message
    return obj(*parameters)


class Parameterisation:
    """Parameterisation specification of the UN object

    args:
        - parm_specification (list): a combo of the distribution family and its parameters; e.g. ['norm', [0, 1]];
        - essence (str): 'distribution' or 'pbox'
    """

    def __init__(self, parm_specification: list, essence: str):
        self.parm_specification = ParamSpecification(
            parm_specification
        )  # combo e.g. ['norm', (0, 1)]
        self.essence = essence

    def yield_construct(self):
        if self.essence == "pbox":
            pbox = match_pbox(
                self.parm_specification.family, self.parm_specification.parameters
            )
            return pbox
        elif self.essence == "distribution":
            dist = pbaDistribution(
                dist_family=self.parm_specification.family,
                dist_params=self.parm_specification.parameters,
            )
            return dist


class ParamSpecification:
    """The combo specification of the distribution family and its parameters

    note:
        Only used for the format of specification.
        a recommended specification: ['gaussian', (12,4)] or ['gaussian', ([0,12],[1,4])]
    """

    def __init__(self, input):
        if (
            not isinstance(input, list)
            or len(input) != 2
            or not isinstance(input[0], str)
        ):
            raise ValueError("Input must be in the format ['str', (a, b)]")

        self.family = input[0]
        self.parameters = input[1]
        # self.un_type_check()

    def supported_distribution_check(self):
        """check if the family is implemented"""
        pass

    def un_type_check(self):
        """infer the real type of UN given the specification
        # NOT in USE. lousy logic
        """

        # distribution case
        if all(isinstance(x, Number) for x in self.parameters):
            self._true_type = "distribution"
        else:
            self._true_type = "pbox"


def pass_down_units(a, b, ops, t):
    """pass down the unit of the uncertain number

    args:
        - a: the first uncertain number
        - b: the second uncertain number
        - ops: the operation to be performed
        - t: the result uncertain number of the operation
    """
    if is_un(b) == 0:
        try:
            new_q = ops(a._physical_quantity, b * a._physical_quantity.units)
        except Exception:
            new_q = ops(a._physical_quantity, b)
    elif is_un(b) == 1:
        new_q = ops(a._physical_quantity, b._physical_quantity)

    t.physical_quantity = new_q
    return t


def is_un(sth):
    """utility function to decide the essence of the object

    returns:
        - 0: if sth is a regular number, float or int
        - 1: if sth is an UncertainNumber object
        - 2: if sth is a construct in {Interval, Pbox, DempsterShafer, or Distribution}
    """

    from ..pba.distributions import Distribution
    from ..pba.dss import DempsterShafer
    from ..pba.pbox_abc import Pbox
    from ..pba.intervals.number import Interval

    if isinstance(sth, Number):
        return 0
    elif isinstance(sth, UncertainNumber):
        return 1
    elif isinstance(sth, Interval | Pbox | DempsterShafer | Distribution):
        return 2


def exist_un(a_list) -> bool:
    """check if there is any UN object in the list"""
    return any(is_un(x) == 1 for x in a_list)


# * ---------------------parametric shortcuts  --------------------- *#


@makeUNPbox
def norm(*args):
    return "norm"


@makeUNPbox
def gaussian(*args):
    return "norm"


@makeUNPbox
def normal(*args):
    return "norm"


@makeUNPbox
def alpha(*args):
    return "alpha"


@makeUNPbox
def anglit(*args):
    return "anglit"


@makeUNPbox
def argus(*args):
    return "argus"


@makeUNPbox
def arcsine(*args):
    return "arcsine"


@makeUNPbox
def beta(*args):
    return "beta"


@makeUNPbox
def betaprime(*args):
    return "betaprime"


@makeUNPbox
def bradford(*args):
    return "bradford"


@makeUNPbox
def burr(*args):
    return "burr"


@makeUNPbox
def burr12(*args):
    return "burr12"


@makeUNPbox
def cauchy(*args):
    return "cauchy"


@makeUNPbox
def chi(*args):
    return "chi"


@makeUNPbox
def chi2(*args):
    return "chi2"


@makeUNPbox
def cosine(*args):
    return "cosine"


@makeUNPbox
def crystalball(*args):
    return "crystalball"


@makeUNPbox
def dgamma(*args):
    return "dgamma"


@makeUNPbox
def dweibull(*args):
    return "dweibull"


@makeUNPbox
def erlang(*args):
    return "erlang"


@makeUNPbox
def expon(*args):
    return "expon"


@makeUNPbox
def exponnorm(*args):
    return "exponnorm"


@makeUNPbox
def exponweib(*args):
    return "exponweib"


@makeUNPbox
def exponpow(*args):
    return "exponpow"


@makeUNPbox
def f(*args):
    return "f"


@makeUNPbox
def fatiguelife(*args):
    return "fatiguelife"


@makeUNPbox
def fisk(*args):
    return "fisk"


@makeUNPbox
def foldcauchy(*args):
    return "foldcauchy"


@makeUNPbox
def genlogistic(*args):
    return "genlogistic"


@makeUNPbox
def gennorm(*args):
    return "gennorm"


@makeUNPbox
def genpareto(*args):
    return "genpareto"


@makeUNPbox
def genexpon(*args):
    return "genexpon"


@makeUNPbox
def genextreme(*args):
    return "genextreme"


@makeUNPbox
def gausshyper(*args):
    return "gausshyper"


@makeUNPbox
def gamma(*args):
    return "gamma"


@makeUNPbox
def gengamma(*args):
    return "gengamma"


@makeUNPbox
def genhalflogistic(*args):
    return "genhalflogistic"


@makeUNPbox
def geninvgauss(*args):
    return "geninvgauss"


@makeUNPbox
def gompertz(*args):
    return "gompertz"


@makeUNPbox
def gumbel_r(*args):
    return "gumbel_r"


@makeUNPbox
def gumbel_l(*args):
    return "gumbel_l"


@makeUNPbox
def halfcauchy(*args):
    return "halfcauchy"


@makeUNPbox
def halflogistic(*args):
    return "halflogistic"


@makeUNPbox
def halfnorm(*args):
    return "halfnorm"


@makeUNPbox
def halfgennorm(*args):
    return "halfgennorm"


@makeUNPbox
def hypsecant(*args):
    return "hypsecant"


@makeUNPbox
def invgamma(*args):
    return "invgamma"


@makeUNPbox
def invgauss(*args):
    return "invgauss"


@makeUNPbox
def invweibull(*args):
    return "invweibull"


@makeUNPbox
def irwinhall(*args):
    return "irwinhall"


@makeUNPbox
def jf_skew_t(*args):
    return "jf_skew_t"


@makeUNPbox
def johnsonsb(*args):
    return "johnsonsb"


@makeUNPbox
def johnsonsu(*args):
    return "johnsonsu"


@makeUNPbox
def kappa4(*args):
    return "kappa4"


@makeUNPbox
def kappa3(*args):
    return "kappa3"


@makeUNPbox
def ksone(*args):
    return "ksone"


@makeUNPbox
def kstwo(*args):
    return "kstwo"


@makeUNPbox
def kstwobign(*args):
    return "kstwobign"


@makeUNPbox
def laplace(*args):

    return "laplace"


@makeUNPbox
def laplace_asymmetric(*args):
    return "laplace_asymmetric"


@makeUNPbox
def levy(*args):
    return "levy"


@makeUNPbox
def levy_l(*args):
    return "levy_l"


@makeUNPbox
def levy_stable(*args):
    return "levy_stable"


@makeUNPbox
def logistic(*args):
    return "logistic"


@makeUNPbox
def loggamma(*args):
    return "loggamma"


@makeUNPbox
def loglaplace(*args):
    return "loglaplace"


@makeUNPbox
def loguniform(*args):
    return "loguniform"


@makeUNPbox
def lomax(*args):
    return "lomax"


@makeUNPbox
def maxwell(*args):
    return "maxwell"


@makeUNPbox
def mielke(*args):
    return "mielke"


@makeUNPbox
def moyal(*args):
    return "moyal"


@makeUNPbox
def nakagami(*args):
    return "nakagami"


@makeUNPbox
def ncx2(*args):
    return "ncx2"


@makeUNPbox
def ncf(*args):
    return "ncf"


@makeUNPbox
def nct(*args):
    return "nct"


@makeUNPbox
def norminvgauss(*args):
    return "norminvgauss"


@makeUNPbox
def pareto(*args):
    return "pareto"


@makeUNPbox
def pearson3(*args):
    return "pearson3"


@makeUNPbox
def powerlaw(*args):
    return "powerlaw"


@makeUNPbox
def powerlognorm(*args):
    return "powerlognorm"


@makeUNPbox
def powernorm(*args):
    return "powernorm"


@makeUNPbox
def rdist(*args):
    return "rdist"


@makeUNPbox
def rayleigh(*args):
    return "rayleigh"


@makeUNPbox
def rel_breitwigner(*args):
    return "rel_breitwigner"


@makeUNPbox
def rice(*args):
    return "rice"


@makeUNPbox
def recipinvgauss(*args):
    return "recipinvgauss"


@makeUNPbox
def semicircular(*args):
    return "semicircular"


@makeUNPbox
def skewcauchy(*args):
    return "skewcauchy"


@makeUNPbox
def skewnorm(*args):
    return "skewnorm"


@makeUNPbox
def studentized_range(*args):
    return "studentized_range"


@makeUNPbox
def t(*args):
    return "t"


@makeUNPbox
def trapezoid(*args):
    return "trapezoid"


@makeUNPbox
def triang(*args):
    return "triang"


@makeUNPbox
def truncexpon(*args):
    return "truncexpon"


@makeUNPbox
def truncpareto(*args):
    return "truncpareto"


@makeUNPbox
def truncweibull_min(*args):
    return "truncweibull_min"


@makeUNPbox
def tukeylambda(*args):
    return "tukeylambda"


def uniform_sps(*args):
    return "uniform"


@makeUNPbox
def vonmises(*args):
    return "vonmises"


@makeUNPbox
def vonmises_line(*args):
    return "vonmises_line"


@makeUNPbox
def wald(*args):
    return "wald"


@makeUNPbox
def weibull_min(*args):
    return "weibull_min"


@makeUNPbox
def weibull_max(*args):
    return "weibull_max"


@makeUNPbox
def wrapcauchy(*args):
    return "wrapcauchy"


# *---------------------some special ones ---------------------*#

from ..pba.pbox_parametric import lognormal, uniform

uniform = constructUN(uniform)
lognormal = constructUN(lognormal)



"""
Solution to the scenario optimization problem for enclosing sets. 

A scenario optimization inputs (1) a table of observations i.e. an (nxd) array, and (2) a shape. 

In general, a scenario program should output three data structures:

(1) A list of integers pointing to the support vectors in the data set;
(2) A data structure representing the set if parametric, e.g. a box, a circle, an ellipse.
(3) A function or object representing the optimal set. For example, the function/method returns true if evaluated inside the set and false otherwise. 

"""

import numpy as numpy
import numpy.matlib as matlib
import scipy.special as sp

def minimal_enclosing_hyperbox(x):
    """
    Inputs
    x: (nxd) array, where n is the size of the data, and d is the number of dimensions of the box

    Outputs
    active_scenarios: list[int], a list of integers pointing to the corresponding active scenarios in the dataset
    box: 2xd array, an enclosing box of dimension d.
    """
    x = numpy.asarray(x,dtype=float) # coherce any iterable to numpy array
    n,d = x.shape
    index,box = [],[]
    for j in range(d):
        s = numpy.argsort(x[:,j])
        index += [s[0],s[-1]]
        box.append([x[s[0],j], x[s[-1],j]])
    active_scenarios = list(set(index)) # list of indices corresponding to the active scenarios
    return active_scenarios, numpy.asarray(box,dtype=float)


def is_inside_box(x,abox):
    """
    x:    (nxd) array
    abox: (dx2) array, or list[list[2 float]] ex.: [[0,1],[3,8],[1,9]] <- a.k.a. interval iterable
    """
    n,d = x.shape
    box = numpy.asarray(abox, dtype=float)
    box_lo = box[:,0]
    box_hi = box[:,1]
    inside = (matlib.repmat(box_lo,n,1) <= x) & (x <= matlib.repmat(box_hi,n,1))
    return numpy.all(inside,axis=1)


triu,tril,log,exp,repmat = numpy.triu, numpy.tril, numpy.log, numpy.exp, numpy.matlib.repmat

def epsLU(k,N,bet):
    """
    Port of the MATLAB code provided by the Authors.

    %% Reference Article 
    % Title Risk and complexity in scenario optimization 
    % Authors S. Garatti and ·M. C. Campi
    % Journal Mathematical Programming 
    % DOI https://doi.org/10.1007/s10107-019-01446-4

    % This function provide the lower and upper reliability parameter for a
    % convex program as defined in Eq (14) of the refernced paper

    % N= number of samples
    % k = Number of scenarios \delta_i for which \zeta_i \geq 0, i.e.  f(x,\delta_i) \geq 0
    % beta = confidence parameter (e.g. very high confidence beta=10^-8)

    """
    alphaL = sp.betaincinv(k,N-k+1,bet)
    alphaU = 1 - sp.betaincinv(N-k+1,k,bet)

    m1 = numpy.asarray(range(k,N+1),dtype=int) # m1 = [k:1:N]
    # ones = numpy.ones((N-k+1,1))
    aux1 = numpy.sum(triu(log(repmat(m1,N-k+1,1)),k=1),axis=1) # aux1 = sum(triu(log(ones(N-k+1,1)*m1),1),2); 
    aux2 = numpy.sum(triu(log(repmat(m1-k,N-k+1,1)),k=1),axis=1) # aux2 = sum(triu(log(ones(N-k+1,1)*(m1-k)),1),2); 
    coeffs1 = aux2-aux1
    m2 = numpy.asarray(range(N+1,4*N+1),dtype=int) # m2 = [N+1:1:4*N]; 
    aux3 = numpy.sum(tril(log(repmat(m2,3*N,1)),k=0),axis=1) # aux3 = sum(tril(log(ones(3*N,1)*m2)),2);
    aux4 = numpy.sum(tril(log(repmat(m2-k,3*N,1)),k=0),axis=1) # aux4 = sum(tril(log(ones(3*N,1)*(m2-k))),2);
    coeffs2 = aux3-aux4
    t1 = 1-alphaL
    t2 = 1
    poly1 = 1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1-(N-m1)*log(t1)))-bet/(6*N) * sum(exp(coeffs2+(m2-N)*log(t1)))
    poly2 = 1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1-(N-m1)*log(t2)))-bet/(6*N) * sum(exp(coeffs2+(m2-N)*log(t2)))

    if (poly1*poly2) > 0:
        epsL = 0
    else:
        while (t2-t1) > 1e-10:
            t = (t1+t2)/2
            polyt = 1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1-(N-m1)*log(t)))-bet/(6*N)*sum(exp(coeffs2+(m2-N)*log(t))); 
            if polyt > 0:
                t1=t
            else:
                t2=t
        epsL = 1-t2
    
    t1 = 0
    t2 = 1-alphaU
    poly1 = 1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1 - (N-m1)*log(t1)))-bet/(6*N)*sum(exp(coeffs2 + (m2-N)*log(t1))); 
    poly2 = 1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1 - (N-m1)*log(t2)))-bet/(6*N)*sum(exp(coeffs2 + (m2-N)*log(t2))); 
    if (poly1*poly2) > 0:
        epsL = 0
    else:
        while (t2-t1) > 1e-10:
            t = (t1+t2)/2
            polyt =1+bet/(2*N)-bet/(2*N)*sum(exp(coeffs1-(N-m1)*log(t)))-bet/(6*N)*sum(exp(coeffs2 + (m2-N)*log(t)))
            if polyt > 0:
                t2=t
            else:
                t1=t
        epsU = 1-t1; 

    return epsL, epsU




if __name__ == '__main__':
    bet=0.1
    k=2
    N=20
    out = epsLU(k,N,bet)
    print(out)




# * ---------------------helper functions  --------------------- *#
from __future__ import annotations
from typing import TYPE_CHECKING
import matplotlib.pyplot as plt
import os

if TYPE_CHECKING:
    from .pba.pbox_abc import Pbox
    from .pba.dss import DempsterShafer


def inspect_un(x: any):
    """Inspect the any type of uncertain number x."""
    print(x.__repr__())
    x.display()


def env_helper(elements: list, env):
    """help visualise the envelope"""

    fig, ax = plt.subplots()

    env.plot(fill_color="salmon", ax=ax, bound_colors=["salmon", "salmon"], zorder=50)

    for p in elements:
        p.plot(ax=ax, zorder=100)


def save_fig(figname, save_dir="."):
    # Ensure directory exists
    os.makedirs(save_dir, exist_ok=True)

    # Build full file path
    filepath = os.path.join(save_dir, f"{figname}.pdf")

    # Save current figure
    plt.savefig(filepath, format="pdf", dpi=300, bbox_inches="tight")



from __future__ import annotations
from typing import TYPE_CHECKING

"""leslie's general bound to bound implementation"""

from ..pba.intervals.number import Interval
from ..pba.intervals.intervalOperators import make_vec_interval
import numpy as np

if TYPE_CHECKING:
    from ..pba.intervals import Interval


def b2b(
    vars: Interval | list[Interval],
    func: callable,
    interval_strategy: str = None,
    subinterval_style: str = None,
    n_sub: int = None,
    n_sam: int = 200,
    **kwargs,
) -> Interval:
    """General implementation of interval propagation through a function:

    .. math::
        Y = g(I_{x1}, I_{x2}, ..., I_{xn})

    where :math:`I_{x1}, I_{x2}, ..., I_{xn}` are intervals.

    In a general case, the function :math:`g` is not necessarily monotonic and :math:`g` may be a black-box model.
    Optimisation to the rescue and two of them particularly: Genetic Algorithm and Bayesian Optimisation.

    args:
        vars (Interval | list[Interval]): a vector Interval or a list or tuple of scalar Intervals, or an EpistemicDomain object, or a scalar Interval;

        func (callable): performance or response function or a black-box model as in subprocess.
            Expect 2D inputs therefore `func` shall have the matrix signature. See Notes for additional details.

        interval_strategy (str): the interval_strategy used for interval propagation. The choice shall be compatible with the response function signature. Seethe notes below for additional details.

            - 'endpoints': only the endpoints

            - 'ga': genetic algorithm

            - 'bo': bayesian optimisation

            - 'direct': directly apply function to the input intervals (the default)

            - 'subinterval': apply function to subintervals

            - 'cauchy_deviate': use the Cauchy Deviate Method for interval propagation

        subinterval_style (str):
            the subinterval_style only used for subinterval propagation, including {''direct'', ''endpoints''}.


        n_sub (int): number of subintervals, only used for subinterval propagation.

        n_sam (int): number of samples, only used for Cauchy deviate method

        **kwargs: additional keyword arguments to be passed to the function, which provides extra control for the algorithm used.
            For example, for GA, one can pass in 'algorithm_param' for BO, one can pass in 'acquisition_function', 'num_iterations' etc.
            Typically, one would try the optimisation harder (e.g. more iterations) for a more accurate result.


    tip:
        This serves as a top-level for generic interval propagation .

    caution:
        ``interval_strategy`` suggests the method of interval propagation (e.g. 'direct' or 'endpoints', or 'subinterval', or 'ga', or 'bo'),  whhile ``subinterval_style`` is only used for 'subinterval' propagation.
        This is to say that what strategy (whether 'direct' or 'endpoints') will be further chosen for the sub-intervals. For scalar function propagation, ``vars`` shuld just be a scalar Interval object rather than a list.

    danger:
        There are some subtleties about the calling signature of the propagating function. For ``endpoints`` strategy/subinterval_style, the function `func` should have a vectorised signature
        as it is expecting a 2D numpy array, whereas for the ``direct`` strategy/subinterval_style, it is expecting to take individual scalar inputs.
        It is recommended to write a function which implements both signature, as seen in the example below.

    returns:
        Interval: the low and upper bound of the response


    example:
        >>> from pyuncertainnumber import b2b
        >>> import numpy as np
        >>> import pyuncertainnumber as pba

        >>> # Define a universal function that handles a performance function with both `vectorised` and `iterable` signatures
        >>> def bar_universal(x):
        ...     if isinstance(x, np.ndarray):  # foo_vectorised signature
        ...         if x.ndim == 1:
        ...             x = x[None, :]
        ...         return x[:, 0] ** 3 + x[:, 1] + 5
        ...     else:
        ...         return x[0] ** 3 + x[1] + 5  # foo_iterable signature

        >>> # Define input intervals
        >>> a = pba.I(3., 5.)
        >>> b = pba.I(6., 26.)

        >>> # using the {'endpoints', "direct", ga", "bo"} strategy
        >>> b2b(vars=[a, b],
        ...     func=bar_universal,
        ...     interval_strategy='endpoints')  # replace with {"direct", ga", "bo"}

        >>> # using the 'subinterval' strategy
        >>> b2b(vars=[a, b],
        ...     func=bar_universal,
        ...     interval_strategy='subinterval',
        ...     subinterval_style='endpoints',
        ...     n_sub=2)

        >>> # using the 'cauchy_deviate' strategy
        >>> b2b(vars=[a, b],
        ...     func=bar_universal,
        ...     interval_strategy='cauchy_deviate',
        ...     n_sam=1000)

        >>> # in comparison, one can compare with the result of interval arithmetic
        >>> def bar_individual(x1, x2):
        ...     return x1 ** 3 + x2 + 5  # individual signature
        >>> bar_individual(a, b)
        [38.0, 156.0]
    """
    from ..pba.intervals.intervalOperators import wc_scalar_interval

    try:
        vec_itvl = make_vec_interval(vars)
    except Exception as e:
        try:
            vec_itvl = wc_scalar_interval(vars)
        except Exception as e:
            raise ValueError(f"Error in making  interval: {e}")

    match interval_strategy:
        case "endpoints":
            return endpoints(vec_itvl, func)
        case "subinterval":
            return subinterval_method(
                vec_itvl,
                func,
                subinterval_style=subinterval_style,
                n_sub=n_sub,
                **kwargs,
            )
        case "ga":
            from ..opt.get_range import get_range_GA
            from pyuncertainnumber import EpistemicDomain

            ep = EpistemicDomain(vec_itvl)
            opt_result = get_range_GA(
                f=func,
                dimension=len(vec_itvl),
                varbound=ep.to_GA_varBounds(),
                verbose=False,
                **kwargs,
            )
            return opt_result[0]  # return the interval only
        case "bo":
            # assumes vectorised func signature
            from ..opt.get_range import get_range_BO
            from pyuncertainnumber import EpistemicDomain

            ep = EpistemicDomain(vec_itvl)

            opt_result = get_range_BO(
                f=func, design_bounds=ep.to_BayesOptBounds(), verbose=False, **kwargs
            )
            return opt_result[0]  # return the interval only
        case "direct":
            return func(vars)
        case (
            "cauchy_deviate"
            | "cauchy_deviate_method"
            | "Cauchy_deviate"
            | "Cauchy_deviate_method"
        ):
            from .cauchy_deviate import cauchy_deviate_method

            y_bound = cauchy_deviate_method(
                input_vector_interval=vec_itvl,
                func=func,
                n_sam=n_sam,
            )
            return y_bound
        case _:
            raise NotImplementedError(
                f"Method {interval_strategy} is not supported yet. Supported methods are: {'endpoints', 'subinterval', 'ga', 'bo', 'direct', 'cauchy_deviate'}"
            )


def endpoints(vec_itvl: Interval, func) -> Interval:
    """Implementation of endpoints method, a.k.a vertex method.

    args:
        vec_itvl (Interval): a vector type Interval object
        func (callable): the function to be evaluated. See notes about function signature.

    note:
        The function `func` is expected to accept a 2D numpy array of shape :math:`(2^d, d)` where `d` is the dimension of the vector interval.
        Therefore, the function  should have a vectorised signature, as opposted to taking individual scalar inputs.


    Example:
        >>> from pyuncertainnumber import pba
        >>> v = pba.I([1, 2], [3, 4])  # a vector interval with two dimensions
        >>> def bar(x): return x[:, 0] ** 3 + x[:, 1] + 5
        >>> endpoints(v, bar)
        Interval(8.0, 36.0)
    """

    v_np = vec_itvl.to_numpy()
    rows = np.vsplit(v_np, v_np.shape[0])
    arr = vec_cartesian_product(*rows)
    # print(arr.shape)  # array of shape (2**n, 2)

    response = func(arr)  # func on each row of combination of endpoints
    min_response = np.min(response)
    max_response = np.max(response)
    return Interval(min_response, max_response)


def subinterval_method(
    vec_itvl: Interval, func, subinterval_style=None, n_sub=None, parallel=False
) -> Interval:
    # TODO parallel subinterval
    """Implmentation of subinterval method which splits subintervals and reconstitutes later.

    Args:
        vec_itvl (Interval): a vector type Interval object

        func (callable): the function to be evaluated. See notes about function signature.

        n_sub (int): number of subintervals

        subinterval_style (str): the subinterval_style used for interval propagation which shall be compatible with the response function signature.

            - 'direct': direct apply function

            - 'endpoints': only the endpoints are propagated

    danger:
        There are some subtleties about function signature. For ``endpoints`` subinterval_style, the function `func` should have a vectorised signature
        as it is expecting a 2D numpy array, whereas for the `direct` subinterval_style, it is expecting to take individual scalar inputs

    Example:
        >>> from pyuncertainnumber import pba
        >>> v = pba.I([1, 2], [3, 4])
        >>> def bar(x): return x[0] ** 3 + x[1] + 5
        >>> b2b.subinterval_method(vec_itvl=v, func=bar, subinterval_style='direct', n_sub=2)
        >>> def bar_vec(x): return x[:, 0] ** 3 + x[:, 1] + 5
        >>> b2b.subinterval_method(vec_itvl=v, func=bar_vec, subinterval_style='endpoints', n_sub=2)
    """
    from pyuncertainnumber.pba.intervals.methods import subintervalise, reconstitute

    if subinterval_style is None:
        raise ValueError(
            "subinterval_style must be chosen within {'direct', 'endpoints'}."
        )
    if n_sub is None:
        raise ValueError("Number of subintervals n_sub must be provided.")

    sub = subintervalise(vec_itvl, n_sub)
    if subinterval_style == "direct":
        row_n = sub.shape[0]
        return reconstitute([func(sub[IND]) for IND in range(row_n)])
    elif subinterval_style == "endpoints":
        return reconstitute([endpoints(sub[i], func) for i in range(len(sub))])


def vec_cartesian_product(*arrays):
    """a vectorised version of the cartesian product

    args:
        arrays: a couple of 1D np.ndarray objects
    """
    grids = np.meshgrid(*arrays, indexing="ij")
    # TODO nice imple but not directrly working with vec Interval object yet
    stacked = np.stack(grids, axis=-1)
    return stacked.reshape(-1, len(arrays))


def i_cartesian_product(a, b):
    """a vectorisation of the interval cartesian product

    todo:
        extend to multiple input arguments
    """
    from pyuncertainnumber import pba

    # Extract bounds
    a_lower = a.lo[:, np.newaxis]  # (2, 1)
    a_upper = a.hi[:, np.newaxis]  # (2, 1)
    b_lower = b.lo[np.newaxis, :]  # (1, 2)
    b_upper = b.hi[np.newaxis, :]  # (1, 2)

    # Broadcast to shape (2, 2)
    cart_lower = np.stack(
        [
            a_lower.repeat(b_lower.shape[1], axis=1),
            np.tile(b_lower, (a_lower.shape[0], 1)),
        ],
        axis=-1,
    )  # shape (2, 2, 2)

    cart_upper = np.stack(
        [
            a_upper.repeat(b_upper.shape[1], axis=1),
            np.tile(b_upper, (a_upper.shape[0], 1)),
        ],
        axis=-1,
    )  # shape (2, 2, 2)

    # Reshape to flat list of interval pairs
    flat_lower = cart_lower.reshape(-1, 2)  # shape (4, 2)
    flat_upper = cart_upper.reshape(-1, 2)  # shape (4, 2)

    return pba.I(lo=flat_lower, hi=flat_upper)


"""
POOL:
            except IndexError as e:
                if "too many indices for array" in str(e):
                    print("2D inputs expected but 1D presented:", e)
                    return endpoints(vec_itvl[None, :], func)
                else:
                    raise  # Re-raise if it's a different IndexError
"""



import enum

""" hint 

Only a pre-set of values are allowed for certain fields in the UncertainNumber class.
"""
# d = Door('open')
# print(d.value)
# d = Door('closed')
# print(d.value)
# d = Door('is not a valid Door')
# print(d.value)

""" ancillary informtion about the uncertain number"""


class AncillaryUncertainty(enum.Enum):
    """
    Base class for all enums that describe different aspects of uncertainty.

    We aim to organise the ancillary information into a list of fields to be
    stored with the quantitative details of the uncertain number rather
    than separately as comments or narrative statements in other files.


    These fields are intended to represent more-or-less orthogonal descriptors that together
    fully characterise what we need to know about the quantity in order to use it in calculations.
    The collection of these fields should be rich enough to account for all relevant and useful
    information, but it is critical that it not seem daunting or burdensome to users. Users are
    never required to enter anything into any of these fields. Numbers that lack substantive
    justifications are indicated with visual cues but, wherever possible, they are not otherwise
    delayed, penalised or restricted in use. Some of the entries for these various fields can be
    populated automatically by computer, and several could be, especially for numbers that arise
    from computing with previously defined numbers.

    The values for the descriptor fields can be given as arbitrary strings with arbitrary formatting
    or by links to other resources. The possible detail for any field is unlimited. For convenience,
    however, suggested entries for many fields can be selected from comprehensive (and
    customisable) dropdown menus.

    Subclasses:
        - Measurand: What is being measured or described (e.g., count, probability, distribution).
        - Ensemble: The collection or grouping over which variability or uncertainty is considered
          (e.g., repeated measurements, flights, households).
        - Variability: The way uncertainty is expressed or summarized
          (e.g., point estimate, confidence).

    """

    pass


# * ----------------------- measurand field


class Measurand(AncillaryUncertainty):
    """What is being measured or described (e.g., count, probability, distribution)."""

    count = "count"
    tally = "tally"
    unobservable_parameter = "unobservable parameter"
    probability = "probability"
    distribution = "distribution"
    range_ = "range"
    rank = "rank"


class Ensemble(AncillaryUncertainty):
    """The collection or grouping over which variability or uncertainty is considered
    (e.g., repeated measurements, flights, households)."""

    repeated_measurements = "repeated measurements"
    flights = "flights"
    pressurisations = "pressurisations"
    temporal_steps = "temporal steps"
    spatial_sites = "spatial sites"
    manufactured_components = "manufactured components"
    customers = "customers"
    people = "people"
    households = "households"
    particular_population = "particular population"


class Variability(AncillaryUncertainty):
    """The way uncertainty is expressed or summarized (e.g., point estimate, confidence)."""

    point_estimate = "point estimate"
    confidence = "confidence"


class Uncertainty_types(AncillaryUncertainty):
    """The type of uncertainty affecting the uncertain number."""

    Certain = "certain"
    Aleatory = "aleatory"
    Epistemic = "epistemic"
    Inferential = "inferential"
    Design_uncertainty = "design uncertainty"
    Vagueness = "vagueness"
    Mixture = "mixture"



"""
------------------------------
cre: Feb 2022

web: github.com/marcodeangelis
org: Univerity of Liverpool

MIT License
------------------------------

These methods are designed to behave neutrally on non-interval inputs.
So, if a non-interval is passed equivalent rules for floats apply.

Interval to float methods, IR -> R:

Interval to bool methods, IR -> {0,1}:

Binary operations, IR2 -> IR
Unary operations, IR -> IR

Parser, R^(nx2) -> IR^n, R^(mxnx2) -> IR^(mxn), R^(2xmxn) -> IR^(mxn)
This method turns an array of compatible dimension into an interval (array).

Subintervalisation methods, IR -> IR^n.

"""

from __future__ import annotations
from typing import Sequence, Sized, Iterable, Optional, Any, Tuple, Union
from itertools import product
import numpy
import numpy as np
from numpy import ndarray, asarray, vstack, linspace, zeros, argmax
from .number import Interval, MACHINE_EPS, lo, hi, width, mid
import json

numpy_min = numpy.min
numpy_max = numpy.max
numpy_sqrt = numpy.sqrt
numpy_abs = numpy.abs
numpy_exp = numpy.exp
numpy_sum = numpy.sum
numpy_sin = numpy.sin
numpy_cos = numpy.cos
numpy_tan = numpy.tan
# numpy_cot = numpy.cotang
numpy_pi = numpy.pi
numpy_inf = numpy.inf
numpy_transpose = numpy.transpose


#####################################################################################
# unary.py
#####################################################################################
# Interval to interval methods. Unary.


def abs(x: Interval):
    """
    Return the absolute value of an Interval.

    If x is not of class Interval, absolute value is returned assuming input is numerical.

    If x is neither a number (neither Interval not numeric), numpy will throw an exception.

    """
    x_lo_abs = numpy.abs(lo(x))
    if is_Interval(x):
        zero_in_x = contain(x, 0)
        x_hi_abs = numpy.abs(hi(x))
        a = numpy.min((x_lo_abs, x_hi_abs), axis=0)
        if x.unsized:
            if zero_in_x:
                a = 0
        else:
            a[zero_in_x] = 0
        b = numpy.max((x_lo_abs, x_hi_abs), axis=0)
        return Interval(a, b)
    return x_lo_abs


def sqrt(x: Interval):
    """
    Return the square root of an Interval.

    If x is not of class Interval, the square root is returned assuming input is numerical.

    If x is neither a number (neither Interval not numeric), numpy will throw an exception.

    """
    x_lo_sqrt = numpy.sqrt(lo(x))
    if is_Interval(x):
        x_hi_sqrt = numpy.sqrt(hi(x))
        return Interval(x_lo_sqrt, x_hi_sqrt)
    return x_lo_sqrt


def exp(x: Interval):
    if is_not_Interval(x):
        return numpy_exp(x)
    return Interval(numpy_exp(lo(x)), numpy_exp(hi(x)))


def log(x: Interval):
    if is_not_Interval(x):
        return numpy.log(x)
    else:
        assert x.lo > 0, "interval has to be positive"
    return Interval(numpy.log(lo(x)), numpy.log(hi(x)))


#####################################################################################
# binary.py
#####################################################################################
# Binary methods between two intervals
# 2-interval to interval. Bianry.


def max(x: Interval, y: Interval):
    if all([is_not_Interval(x), is_not_Interval(y)]):
        return numpy.max((x, y), axis=0)
    a = numpy.max((lo(x), lo(y)), axis=0)
    b = numpy.max((hi(x), hi(y)), axis=0)
    return Interval(a, b)


def min(x: Interval, y: Interval):
    if all([is_not_Interval(x), is_not_Interval(y)]):
        return numpy.min((x, y), axis=0)
    a = numpy.min((lo(x), lo(y)), axis=0)
    b = numpy.min((hi(x), hi(y)), axis=0)
    return Interval(a, b)


def env(x: Interval, y: Interval):
    """Return the envelope interval containing both x and y."""
    if all([is_not_Interval(x), is_not_Interval(y)]):
        return numpy.array([numpy.min((x, y)), numpy.max((x, y))])
    a = numpy.min((lo(x), lo(y)), axis=0)
    b = numpy.max((hi(x), hi(y)), axis=0)
    return Interval(a, b)


#####################################################################################
# trig.py
#####################################################################################


def sin(x: Interval):
    """
    Implementation of Interval Arithmetic in CORA 2016

    Matthias Althoff and Dmitry Grebenyuk

    EPiC Series in Computing Volume 43, 2017, Pages 91-105

    ARCH16. 3rd International Workshop on Applied Verification for Continuous and Hybrid Systems
    """
    if not (is_Interval(x)):
        return numpy_sin(x)  # int, float, ndarray

    if not (x.scalar):
        return sin_vector(x)

    twopi = 2 * numpy_pi
    pihalf = numpy_pi / 2

    if width(x) >= twopi:
        return Interval(-1, 1)

    domain1 = Interval(0, pihalf)
    domain2 = Interval(pihalf, 3 * pihalf)
    domain3 = Interval(3 * pihalf, twopi)

    yl = x.lo % twopi
    yh = x.hi % twopi
    y = Interval(lo=yl, hi=yh)

    sin_l = numpy_sin(yl)
    sin_h = numpy_sin(yh)

    if contain(domain1, y) & (yl <= yh):
        return Interval(sin_l, sin_h)
    if contain(domain2, y) & (yl <= yh):
        return Interval(sin_h, sin_l)
    if contain(domain3, y) & (yl <= yh):
        return Interval(sin_l, sin_h)

    case1a = contain(domain1, yl) & contain(domain1, yh) & (yl > yh)
    case1b = contain(domain1, yl) & contain(domain3, yh)
    case1c = contain(domain2, yl) & contain(domain2, yh) & (yl > yh)
    case1d = contain(domain3, yl) & contain(domain3, yh) & (yl > yh)

    case2a = contain(domain1, yl) & contain(domain1, yh) & (yl <= yh)
    case2b = contain(domain3, yl) & contain(domain1, yh)
    case2c = contain(domain3, yl) & contain(domain3, yh) & (yl <= yh)

    case3a = contain(domain1, yl) & contain(domain2, yh)
    case3b = contain(domain3, yl) & contain(domain2, yh)

    case4a = contain(domain2, yl) & contain(domain1, yh)
    case4b = contain(domain2, yl) & contain(domain3, yh)

    case5 = contain(domain2, yl) & contain(domain2, yh) & (yl <= yh)

    if case1a | case1b | case1c | case1d:
        return Interval(-1, 1)
    if case2a | case2b | case2c:
        return Interval(sin_l, sin_h)
    if case3a | case3b:
        return Interval(min(sin_l, sin_h), 1)
    if case4a | case4b:
        return Interval(-1, max(sin_l, sin_h))
    if case5:
        return Interval(sin_h, sin_l)


def sin_vector(x: Interval):  # vectorised version of sin().

    if x.unsized:
        return sin(x)

    twopi = 2 * numpy_pi
    pihalf = numpy_pi / 2

    mask1a = width(x) >= twopi

    domain1 = Interval(0, pihalf)
    domain2 = Interval(pihalf, 3 * pihalf)
    domain3 = Interval(3 * pihalf, twopi)

    yl = x.lo % twopi
    yh = x.hi % twopi
    y = Interval(yl, yh)

    sin_l = numpy_sin(yl)
    sin_h = numpy_sin(yh)

    # [l,h] all else
    a = sin_l.copy()
    b = sin_h.copy()

    # [-1,1]
    mask3a = contain(domain1, yl) & contain(domain1, yh) & (yl > yh)
    mask3b = contain(domain1, yl) & contain(domain3, yh)
    mask3c = contain(domain2, yl) & contain(domain2, yh) & (yl > yh)
    mask3d = contain(domain3, yl) & contain(domain3, yh) & (yl > yh)
    case1 = mask1a | mask3a | mask3b | mask3c | mask3d
    a[case1] = -1
    b[case1] = 1
    if all(case1):
        return Interval(lo=a, hi=b)
    # [h,l]
    mask2b = (
        contain(domain2, yl[~case1])
        & contain(domain2, yh[~case1])
        & (yl[~case1] <= yh[~case1])
    )  # return Interval(sin_h,sin_l)
    case2 = mask2b
    a[case2] = sin_h[case2]
    b[case2] = sin_l[case2]
    # [min, 1]
    mask5a = contain(domain1, yl[~case1]) & contain(domain2, yh[~case1])
    mask5b = contain(domain3, yl[~case1]) & contain(domain2, yh[~case1])
    case3 = mask5a | mask5b
    a[case3] = min(sin_l[case3], sin_h[case3])
    b[case3] = 1
    # [-1, max]
    mask6a = contain(domain2, yl[~case1]) & contain(domain1, yh[~case1])
    mask6b = contain(domain2, yl[~case1]) & contain(domain3, yh[~case1])
    case4 = mask6a | mask6b
    a[case4] = -1
    b[case4] = max(sin_l[case4], sin_h[case4])
    return Interval(lo=a, hi=b)


def cos(x: Interval):
    """
    Implementation of Interval Arithmetic in CORA 2016

    Matthias Althoff and Dmitry Grebenyuk

    EPiC Series in Computing Volume 43, 2017, Pages 91-105

    ARCH16. 3rd International Workshop on Applied Verification for Continuous and Hybrid Systems
    """
    if not (is_Interval(x)):
        return numpy_cos(x)  # int, float, ndarray

    if not (x.scalar):
        return cos_vector(x)

    twopi = 2 * numpy_pi

    # [-1,1] aka case 0
    if width(x) >= twopi:
        return Interval(-1, 1)

    domain1 = Interval(0, numpy_pi)
    domain2 = Interval(numpy_pi, 2 * numpy_pi)

    yl = x.lo % twopi
    yh = x.hi % twopi
    y = Interval(lo=yl, hi=yh)

    cos_l = numpy_cos(yl)
    cos_h = numpy_cos(yh)

    # [-1,1]
    case1a = (yh < yl) & contain(domain1, yl) & contain(domain1, yh)
    case1b = (yh < yl) & contain(domain2, yl) & contain(domain2, yh)
    # [cos_l, cos_h]
    case2a = (yl <= yh) & contain(domain2, yl) & contain(domain2, yh)
    # [min(cos_l, cos_h), 1]
    case3a = contain(domain2, yl) & contain(domain1, yh)
    # [-1, max(cos_l, cos_h)]
    case4a = contain(domain1, yl) & contain(domain2, yh)
    # [cos_h, cos_l]
    case5a = (yl <= yh) & contain(domain1, yl) & contain(domain1, yh)

    if case1a | case1b:
        return Interval(-1, 1)
    if case2a:
        return Interval(cos_l, cos_h)
    if case3a:
        return Interval(min(cos_l, cos_h), 1)
    if case4a:
        return Interval(-1, max(cos_l, cos_h))
    if case5a:
        return Interval(cos_h, cos_l)


def cos_vector(x: Interval):  # vectorised version of cos()
    if x.unsized:
        return sin(x)

    twopi = 2 * numpy_pi

    case0 = width(x) >= twopi

    domain1 = Interval(0, numpy_pi)
    domain2 = Interval(numpy_pi, 2 * numpy_pi)

    yl = x.lo % twopi
    yh = x.hi % twopi

    cos_l = numpy_cos(yl)
    cos_h = numpy_cos(yh)

    a = cos_l.copy()
    b = cos_h.copy()

    # [-1,1]
    case1a = (yh < yl) & contain(domain1, yl) & contain(domain1, yh)
    case1b = (yh < yl) & contain(domain2, yl) & contain(domain2, yh)
    case1 = case0 | case1a | case1b
    a[case1] = -1
    b[case1] = 1
    # [cos_l, cos_h]
    # case2 = (yl<=yh) & contain(domain2,yl) & contain(domain2,yh)
    # a[case2] = cos_l[case2]
    # b[case2] = cos_h[case2]
    # [min(cos_l, cos_h), 1]
    case3 = contain(domain2, yl) & contain(domain1, yh)
    a[case3] = min(cos_l[case3], cos_h[case3])
    b[case3] = 1
    # [-1, max(cos_l, cos_h)]
    case4 = contain(domain1, yl) & contain(domain2, yh)
    a[case4] = -1
    b[case4] = max(cos_l[case4], cos_h[case4])
    # [cos_h, cos_l]
    case5 = (yl <= yh) & contain(domain1, yl) & contain(domain1, yh)
    a[case5] = cos_h[case5]
    b[case5] = cos_l[case5]
    return Interval(lo=a, hi=b)


def tan(x: Interval):
    """
    Implementation of Interval Arithmetic in CORA 2016

    Matthias Althoff and Dmitry Grebenyuk

    EPiC Series in Computing Volume 43, 2017, Pages 91-105

    ARCH16. 3rd International Workshop on Applied Verification for Continuous and Hybrid Systems
    """

    if not (is_Interval(x)):
        return numpy_tan(x)  # int, float, ndarray

    if not (x.scalar):
        return tan_vector(x)

    pihalf = numpy_pi / 2

    domain1 = Interval(0, pihalf)
    domain2 = Interval(pihalf, numpy_pi)

    zl = x.lo % numpy_pi
    zh = x.hi % numpy_pi

    # [-∞, ∞]
    case1a = width(x) > numpy_pi
    case1b = (zh < zl) & contain(domain1, zl) & contain(domain1, zh)
    case1c = (zh < zl) & contain(domain2, zl) & contain(domain2, zh)
    case1d = contain(domain1, zl) & contain(domain2, zh)

    # [tan_l, tan_h]
    case2a = (zl <= zh) & contain(domain1, zl) & contain(domain1, zh)
    case2b = (zl <= zh) & contain(domain2, zl) & contain(domain2, zh)

    if case1a | case1b | case1c | case1d:
        return Interval(-numpy_inf, numpy_inf)
    if case2a | case2b:
        return Interval(tan(zl), tan(zh))
    else:
        return Interval(tan(zl), tan(zh))


def tan_vector(x: Interval):  # Vectorised version of tan().
    if x.unsized:
        return tan(x)

    pihalf = numpy_pi / 2

    domain1 = Interval(0, pihalf)
    domain2 = Interval(pihalf, numpy_pi)

    zl = x.lo % numpy_pi
    zh = x.hi % numpy_pi

    tan_l = tan(zl)
    tan_h = tan(zh)

    a = tan_l.copy()
    b = tan_h.copy()

    # [-∞, ∞]
    case1a = width(x) > numpy_pi
    case1b = (zh < zl) & contain(domain1, zl) & contain(domain1, zh)
    case1c = (zh < zl) & contain(domain2, zl) & contain(domain2, zh)
    case1d = contain(domain1, zl) & contain(domain2, zh)
    case1 = case1a | case1b | case1c | case1d
    a[case1] = -numpy_inf
    b[case1] = numpy_inf

    # #[tan_l, tan_h]
    # case2a = (zl<=zh) & contain(domain1,zl) & contain(domain1,zh)
    # case2b = (zl<=zh) & contain(domain2,zl) & contain(domain2,zh)
    return Interval(lo=a, hi=b)


# Interval to ndarray[float]
# def linspace():


# def union(): pass # spell the difference between union and subpaving.
# def intersection():
# def difference():
# def set_difference(x:Interval,y:Interval):


#####################################################################################
# set.py
#####################################################################################
# Interval to bool methods, Binary.
# ...
def straddle_zero(x: Interval) -> bool:
    if x.unsized:
        return (lo(x) <= 0) & (hi(x) >= 0)
    else:
        return any((lo(x).flatten() <= 0) & (hi(x).flatten() >= 0))


def intersect(x: Interval, y: Interval):
    return ~((x < y) | (y < x))  # commutative


def contain(x: Interval, y: Interval):
    return (lo(x) <= lo(y)) & (hi(x) >= hi(y))  # x contain y


def almost_contain(x: Interval, y: Interval, tol=1e-9):
    return (
        lo(y) -
        # x contain y
        lo(x)
        > -tol
    ) & (hi(x) - hi(y) > -tol)


def intersect_vector(x_: Interval, y_: Interval):
    """
    This function checks if the focal elements x, intersect the subpaving y.

    x: A n-list of d-boxes or d-intervals, e.g. a subpaving. x.shape=(r,d)
    y: A m-list of d-boxes or d-intervals, e.g. a focal element. y.shape=(p,d)

    out: A (rp)-list of d-arrays of booleans
    """
    x = intervalise(x_)
    y = intervalise(y_)
    # n,d = x.shape
    m, d = y.shape
    x_lo = lo(x)  # x_lo = numpy.array([xi.lo for xi in x])
    x_hi = hi(x)  # x_hi = numpy.array([xi.hi for xi in x])
    where_intersect = numpy.zeros((m,), dtype=bool)  # inter = []
    for i, yi in enumerate(tolist(y)):  #  a focal elem
        where_intersect[i] = any(
            numpy.all(~((x_hi < lo(yi)) | (hi(yi) < x_lo)), axis=1)
        )
    return where_intersect


#####################################################################################
# parser.py
#####################################################################################
# Universal parser


def intervalise(x_: Any, interval_index=-1) -> Union[Interval, Any]:
    """
    This function casts an array-like structure into an Interval structure.
    All array-like structures will be first coerced into an ndarray of floats.
    If the coercion is unsuccessful the following error is thrown: `ValueError: setting an array element with a sequence.`

    For example this is the expected behaviour:
    (*) an ndarray of shape (4,2) will be cast as an Interval of shape (4,).

    (*) an ndarray of shape (7,3,2) will be cast as an Interval of shape (7,3).

    (*) an ndarray of shape (3,2,7) will be cast as a degenerate Interval of shape (3,2,7).

    (*) an ndarray of shape (2,3,7) will be cast as an Interval of shape (3,7).

    (*) an ndarray of shape (2,3,7,2) will be cast as an Interval of shape (2,3,7) if interval_index is set to -1.

    If an ndarray has shape with multiple dimensions having size 2, then the last dimension is intervalised.
    So, an ndarray of shape (7,2,2) will be cast as an Interval of shape (7,2) with the last dimension intervalised.
    When the ndarray has shape (2,2) again is the last dimension that gets intervalised.

    In case of ambiguity, e.g. (2,5,2), now the first dimension can be forced to be intervalised, selecting index=0, default is -1.

    It returns an interval only if the input is an array-like structure, otherwise it returns the following numpy error:
    `ValueError: setting an array element with a sequence.`

    TODO: Parse a list of mixed numbers: interval and ndarrays.

    """

    def treat_list(xx):
        xi_lo, xi_hi = [], []
        for xi in xx:  # if each element in the list is an interval of homogeneus shape
            if xi.__class__.__name__ == "Interval":
                xi_lo.append(xi.lo)
                xi_hi.append(xi.hi)
            else:
                xi_ = intervalise(xi)  # recursion
                xi_lo.append(xi_.lo)
                xi_hi.append(xi_.hi)
        try:
            return Interval(lo=xi_lo, hi=xi_hi)
        except:
            print("!! Parsing an interval from list failed.")
            return x_

    if x_.__class__.__name__ == "Interval":
        return x_
    try:
        x = asarray(x_, dtype=float)
    # ValueError: setting an array element with a sequence. The requested array has an inhomogeneous shape after 1 dimensions. The detected shape was (...) + inhomogeneous part.
    except ValueError:
        if x_.__class__.__name__ == "list":
            # attempt to turn a n-list of intervals into a (n,...)-interval
            return treat_list(x_)
    s = x.shape
    two = [si == 2 for si in s]
    # if all(two): return Interval(lo=numpy_transpose(x)[0],hi=numpy_transpose(x)[1])
    if all(two):
        return Interval(lo=x[..., 0], hi=x[..., 1])
    elif any(two):
        # if two[-1]: return Interval(lo=numpy_transpose(x)[0],hi=numpy_transpose(x)[1]) # the last dimension has size 2
        if two[-1]:
            # last dimension has size 2
            return Interval(lo=x[..., 0], hi=x[..., 1])
        elif two[0]:
            return Interval(lo=x[0], hi=x[1])  # first dimension has size 2
        elif (two[-1]) & (two[0]):  # this is the ambiguous case (2,3,5,2)
            if interval_index == 0:
                # first dimension gets intervalised
                return Interval(lo=x[0], hi=x[1])
            # elif index == -1: return Interval(lo=numpy_transpose(x)[0],hi=numpy_transpose(x)[1])
            elif interval_index == -1:
                return Interval(lo=x[..., 0], hi=x[..., 1])
            # if (sum(two)==1) & (two[0]): return Interval(lo=x[0],hi=x[1])# there is only one dimension of size 2 and is the first one
        print(
            "Array-like structure must have the last (or first) dimension of size 2, for it to be coerced to Interval."
        )
        return Interval(lo=x)
    else:
        return Interval(lo=x)


def sizeit(x: Interval) -> Interval:
    """
    Takes an unsized scalar interval and turns it in to a sized one.
    """
    if is_Interval(x):
        if x.scalar & x.unsized:
            return Interval(lo=[x.lo], hi=[x.hi])
    return x


def unsizeit(x: Interval) -> Interval:
    """
    Takes a sized scalar interval and turns it in to a unsized one.

    """
    if is_Interval(x):
        if x.scalar & x.unsized == False:
            return Interval(lo=x.lo[0], hi=x.hi[0])
    return x


def tolist(x: Interval):
    if is_not_Interval(x):
        return x
    dim = len(x.shape)
    if dim == 1:
        return [Interval(xi.lo, xi.hi) for xi in x]
    if dim == 2:
        m, d = x.shape
        return [Interval(lo=x.lo[i, :], hi=x.hi[i, :]) for i in range(m)]
    if dim > 2:
        return x  # not implemented yet


#####################################################################################
# subint.py
#####################################################################################


def subintervalise(x_: Interval, n: Union[int, tuple] = 0) -> Interval:
    """
    return:
        - matrix Interval
    """
    x = intervalise(x_)
    d = len(x.shape)  # dimension of the array
    if n == 0 | n == 1:
        return x  # should return a subtiling (sized interval)
    if x.scalar:  # or x.scalar == True
        xx = linspace(x.lo, x.hi, num=n + 1)
        return intervalise(vstack([xx[:-1], xx[1:]]))
    elif d == 1:  # x.shape = (m,)
        m = x.shape[0]  # size of 1d array
        if type(n) == int:
            n = m * [n]  # differential split number
        X_sub = []
        for i, xi in enumerate(x):
            xxi = subintervalise(xi, n=n[i])  # recursion
            X_sub.append(sizeit(xxi).val)
        return intervalise(asarray(list(product(*X_sub)), dtype=float))
    #     elif len(x.shape)>1: pass # TODO: implement space-product subintervalization with arrays of dimension greater than 2.
    else:
        print(
            "!! Subtiling not yet supported for interval arrays of dimension 2 or larger. Input will be returned."
        )
    return x


def split_interval(x: Interval, y: float = None):
    if y is None:
        return Interval(lo(x), mid(x)), Interval(mid(x), hi(x))

    x1, x2 = Interval(lo(x), hi(x)), Interval(
        lo(x), hi(x)
    )  # TODO: implement copy method
    y_in_x = contain(x, y)  # x contain y
    if x.unsized:
        if ~y_in_x:
            return x1, x2
        return Interval(lo=lo(x), hi=y), Interval(lo=y, hi=hi(x))
    else:
        pass
        # x1[y_in_x] = Interval(lo(x)[y_in_x],hi=y)
    # x2[y_in_x] = Interval(y,hi=hi(x)[y_in_x])
    return x1, x2


def reconstitute(x_: Interval):
    x = intervalise(x_)
    d = len(x.shape)  #  dimension of the subtiling ==1 if scalar, ==2 if 1d array
    if d == 1:
        return Interval(lo=numpy.min(x.lo), hi=numpy.max(x.hi))
    elif d == 2:
        return Interval(lo=numpy.min(x.lo, axis=1), hi=numpy.max(x.hi, axis=1))
    else:
        print(
            "!! Subtiling not yet supported for interval arrays of dimension 2 or larger."
        )
    return x


def space_product(x_: Union[ndarray, Interval], y_: Union[ndarray, Interval]):
    return asarray(tuple(product(x_, y_)))


def bisect(x_: Interval, i: int = None):
    """
    :x_: Interval of shape (n,)

    Bisect the largest box if i is None.
    """
    x = intervalise(x_)
    if x.scalar:
        mid_x = mid(x)
        return Interval(lo(x), mid_x), Interval(mid_x, hi(x))
    if i is not None:
        split_index = i
    else:
        w = width(x)
        split_index = argmax(w)
    d = x.shape[0]
    n = [0] * d
    # ex: (0,0,2,0,0,0) if interval of dim 6 has third dimension bisected
    n[split_index] = 2
    x_bisect = subintervalise(x, n=tuple(n))
    return x_bisect[0], x_bisect[1]


#####################################################################################
# types.py
#####################################################################################
# Interval to bool methods, Unary.


def is_Interval(x: Any) -> bool:
    return x.__class__.__name__ == "Interval"


def is_not_Interval(x: Any) -> bool:
    return x.__class__.__name__ != "Interval"


#####################################################################################
################################# neural_networks.py ################################
#####################################################################################


def dot(x: Interval, y: Interval):
    return sum(x * y)


def rowcol_old(W, x):
    """
    (m,n) x (n,1) -> (m,1)
    (m,n) x (n,p) -> (m,p)
    (1,n) x (n,1) -> (1,1)
    """
    s = W.shape
    x_shape = x.shape
    if x_shape[0] == 1:  # x is row and must be either squeezed or transposed
        x_ = x[0, :]
    if x_shape[1] == 1:  # this is the correct shape
        x_ = x[:, 0]
    if len(x_shape) == 1:  # x can be row or column (n,)
        x_ = x
    y = []
    for i in range(s[0]):
        y.append(dot(W[i], x_))
    return intervalise(y)


def rowcol_W_x(W, x):
    """
    Row by column multiplication between a matrix W and a column vector x.

    (m,n) x (n,1) -> (m,1)
    (1,n) x (n,1) -> (1,1)
    The following cases are also accepted even though mathematically impossible
    (m,n) x (n,) -> (m,1)
    (1,n) x (n,1) -> (1,1)
    (1,n) x (1,n) -> (1,1)
    """
    m, n = W.shape
    x_shape = x.shape
    if not ((n == x_shape[0]) | (n == x_shape[1])):
        raise ValueError(
            f"Incompatible shapes [{m}, {n}] x({x_shape[0]}, 1) -> (?, ?). Inner sizes must be same, {x_shape[1]} is different from {n}."
        )
    if x_shape[0] == 1:  # x is row and must be either squeezed or transposed
        x_ = x[0, :]
    if x_shape[1] == 1:  # this is the correct shape
        x_ = x[:, 0]
    if len(x_shape) == 1:  # x can be row or column (n,)
        x_ = x
    y = numpy.empty((m, 2))
    for i in range(m):
        inner_product = dot(W[i], x_)
        y[i, 0] = inner_product.lo
        y[i, 1] = inner_product.hi
    ylo = numpy.expand_dims(y[..., 0], axis=1)
    yhi = numpy.expand_dims(y[..., 1], axis=1)
    return Interval(ylo, yhi)


def rowcol_xT_WT(x, W):
    """
    Row by column multiplication between the row vector xT and the matrix transpose WT.
    (1,n) x (n,m) -> (1,m)
    (1,n) x (n,1) -> (1,1)
    The following cases are also accepted even though mathematically impossible
    (,n) x (n,m) -> (1,m)
    (n,1) x (n,m) -> (1,1)
    """
    n, m = W.shape
    x_shape = x.shape
    if not ((n == x_shape[0]) | (n == x_shape[1])):
        raise ValueError(
            f"Incompatible shapes(1, {x_shape[1]}) x({n}, {m}) -> (?, ?). Inner sizes must be same, {x_shape[1]} is different from {n}."
        )
    if x_shape[0] == 1:  # this is the correct shape
        x_ = x[0, :]
    if x_shape[1] == 1:  # x is row and must be either squeezed or transposed
        x_ = x[:, 0]
    if len(x_shape) == 1:  # x can be row or column (n,)
        x_ = x
    y = numpy.empty((m, 2))
    for i in range(m):
        inner_product = dot(x_, W[..., i])
        y[i, 0] = inner_product.lo
        y[i, 1] = inner_product.hi
    ylo = numpy.expand_dims(y[..., 0], axis=0)
    yhi = numpy.expand_dims(y[..., 1], axis=0)
    return Interval(ylo, yhi)


def matmul(A, B):
    """
    (m,n) x (n,p) -> (m,p)
    (1,n) x (n,1) -> (1,1)
    """
    m, na = A.shape
    nb, p = B.shape
    if na != nb:
        raise ValueError(
            f"Incompatible shapes({m}, {na}) x({nb}, {p}) -> (?, ?). Inner sizes must be same, {na} is different from {nb}."
        )
    C = numpy.empty((m, p, 2))
    for i in range(m):
        for j in range(p):
            inner_product = dot(A[i], B[..., j])
            C[i, j, 0] = inner_product.lo
            C[i, j, 1] = inner_product.hi
    return intervalise(C)


def transpose(x: Interval):  # not efficient it creates a new object in memory
    """
    Input an interval of shape (m,n) returns an interval of shape (n,m).
    """
    return Interval(x.val[..., 0], x.val[..., 1])


def squeeze(x: Interval):  # not efficient it creates a new object in memory
    return Interval(numpy.squeeze(x.lo), numpy.squeeze(x.hi))


################################ activation_functions.py #############################


def relu_nointerval(x: ndarray):
    positive = x > 0
    output = numpy.zeros(x.shape)
    output[positive] = x[positive]
    return output


def relu_deriv(x: ndarray):
    positive = x > 0
    output = numpy.zeros(x.shape)
    output[positive] = 1
    return output


def relu(x: Interval):
    if is_not_Interval(x):
        return relu_nointerval(x)
    case_1 = x.hi < 0
    x_lo = x.val[..., 0].T
    x_hi = x.val[..., 1].T
    x_lo[case_1] = 0
    x_hi[case_1] = 0
    case_3 = (x_lo < 0) & (numpy.logical_not(case_1))
    x_lo[case_3] = 0
    relu_x = Interval(lo=x_lo, hi=x_hi)
    return relu_x


def relu_deriv_interval(x: Interval):
    if is_not_Interval(x):
        return relu_deriv(x)
    x_lo = x.val[..., 0].T
    x_hi = x.val[..., 1].T
    case_1 = x_hi < 0
    d_lo = numpy.zeros(x_lo.shape)
    d_hi = numpy.ones(x_hi.shape)
    d_hi[case_1] = 0
    # case_3 = (x_lo<0) & (np.logical_not(case_1))
    case_2 = x_lo > 0
    d_lo[case_2] = 1
    d_relu_x = Interval(lo=d_lo, hi=d_hi)
    return d_relu_x


def sigmoid(x):
    return 1 / (1 + exp(-x))


def sigmoid_deriv(x):
    return sigmoid(x) * (1 - sigmoid(x))


# def tanh_(x): return np.tanh(x)


def tanh(x: Interval):
    return (exp(2 * x) - 1) / (exp(2 * x) + 1)


def tanh(x: Interval):
    r = -1
    s = 1
    u = 1
    t = 1
    return s / u - (s * t / u**2 - r / u) / (t / u + exp(2 * x))


# def cot(x): return 1/np.tan(x)
# def tanh(x): return -(1/(cot(np.arctan(x)/2)**2))
def cosh(x: Interval):
    return (1 + exp(-2 * x)) / (2 * exp(-x))


def tanh_deriv(x: Interval):
    return (1 / cosh(x)) ** 2


def unpack(x: Interval) -> list[Interval]:
    """unpack an array-like Interval object into a list of item intervals

    return:
        a list of scalar Interval objects

    example:
        >>> x = Interval(lo=([0, 1]), hi=([1, 2]))
        >>> unpack(x)
        [Interval(0, 1), Interval(1, 2)]
    """

    if x.is_scalar:
        raise ValueError("a scalar interval cannot be unpacked.")

    if len(x) >= 0:
        return [item for item in x]


# * ----------------------------------- reload from json
def load_interval_from_json(filename: str) -> Interval:
    """Load a NumPy array from a JSON file saved by save_array_to_json().

    note:
        Both .json and .json5 files are supported.

    example:
        >>> interval = load_interval_from_json("interval_data.json5")
    """
    import json5

    with open(filename, "r") as f:
        data = json5.load(f)
    array = np.array(data, dtype=float)
    the_I = intervalise(array)
    return the_I



from __future__ import annotations
from typing import TYPE_CHECKING
from numbers import Number
import numpy as np

""" This module contains the some examplar performance/response functions  """


def func_inspect():
    """inspect the formatting of the user-defined function"""
    pass


if TYPE_CHECKING:
    from ..pba.intervals.number import Interval


# * ------------- some of the function for testing APIs


def foo_args(x0, x1, x2):
    """func signature with individual inputs (arguments)"""
    return x0**3 + x1 + x2


def foo_iter(x):
    """func signature with iterable input (list, tuple, np.ndarray)"""
    return x[0] ** 3 + x[1] + x[2]


def foo_vec(x):
    """func signature with matrix input (2D np.ndarray)"""
    return x[:, 0] ** 3 + x[:, 1] + x[:, 2]


def foo_universal(x):
    """a universal signature that takes iterable and matrix inputs but not individual inputs

    caveat:
        - this function works with a vector Interval object
        - but wrong answer with a 2d matrix Interval object due to unclear broadcasting rules
    """

    if isinstance(x, np.ndarray):  # foo_vectorised signature
        if x.ndim == 1:
            x = x[None, :]
        return x[:, 0] ** 3 + x[:, 1] + x[:, 2]
    else:
        return x[0] ** 3 + x[1] + x[2]  # foo_iterable signature


def bar(x):
    """a simple 2-dimensional function for testing purposes"""
    return x[0] ** 3 + x[1] + 5


def bar_vec(x):
    """a simple vectroised 2-dimensional function for testing purposes"""
    return x[:, 0] ** 3 + x[:, 1] + 5


def bar_universal(x):
    """a universal signature that takes iterable and matrix inputs but not individual inputs

    caveat:
        - this function works with a vector Interval object
        - but wrong answer with a 2d matrix Interval object due to unclear broadcasting rules
    """

    if isinstance(x, np.ndarray):
        if x.ndim == 1:
            x = x[None, :]
        return x[:, 0] ** 3 + x[:, 1] + 5  # vectorised signature
    else:
        return x[0] ** 3 + x[1] + 5  # iterable signature


def bar_individual(x0, x1):
    """a simple 2-dimensional function for testing purposes with individual inputs"""
    return x0**3 + x1 + 5


#! 'func' needs to take 2D inputs for maxmising the potential for array computation
def cb_func(x):
    """Calculates deflection and stress for a cantilever beam.

    Args:
        x (np.array): Array of input parameters:
            x[0]: Distance from the neutral axis to the point of interest (m)
            x[1]: Length of the beam (m)
            x[2]: Second moment of area (mm^4)
            x[3]: Applied force (N)
            x[4]: Young's modulus (MPa)

    Returns:
        np.array([deflection (m), stress (MPa)])
               Returns np.array([np.nan, np.nan]) if calculation error occurs.
    """

    y = x[0]
    beam_length = x[1]
    I = x[2]
    F = x[3]
    E = x[4]
    try:  # try is used to account for cases where the input combinations leads to error in fun due to bugs
        deflection = F * beam_length**3 / (3 * E * 10**6 * I)  # deflection in m
        stress = F * beam_length * y / I / 1000  # stress in MPa

    except:
        deflection = np.nan
        stress = np.nan

    return np.array([deflection, stress])


def cb_deflection(x: list[Number] | np.ndarray | Interval | list[Interval]):
    """Calculates deflection and stress for a cantilever beam.

    Args:
        x (np.array): Array of input parameters:
            x[0]: Length of the beam (m)
            x[1]: Second moment of area (mm^4)
            x[2]: Applied force (N)
            x[3]: Young's modulus (MPa)

    Returns:
        float: deflection (m)
               Returns np.nan if calculation error occurs.
    """

    beam_length = x[0]
    I = x[1]
    F = x[2]
    E = x[3]
    try:  # try is used to account for cases where the input combinations leads to error in fun due to bugs
        deflection = F * beam_length**3 / (3 * E * 10**6 * I)  # deflection in m

    except:
        deflection = np.nan

    return deflection


def cb_deflection_vectorised(x: np.ndarray | Interval):
    """Vectorised version for calculating cantilever beam deflection over multiple inputs.

    args:
        x (Interval or np.ndarray) : 2D inputs of  matrix shape, x.shape = (N, 4)
    """

    #! code below works great for np arrays but not for interval objects
    # x = np.atleast_2d(x)  # Ensures x is at least 2D: (N, 4)

    if x.ndim == 1:
        return cb_deflection(x)

    beam_length = x[:, 0]
    I = x[:, 1]
    F = x[:, 2]
    E = x[:, 3]
    deflection = F * beam_length**3 / (3 * E * 10**6 * I)

    if deflection.shape[0] == 1:
        return deflection[0]  # Return scalar
    return deflection


def cb_deflection_separate(L, I, F, E):
    """compute the deflection in the cantilever beam example

    # TODO add typing for UncertainNumber
    Args:
        L (UncertainNumber): Length of the beam (m)
        I: Second moment of area (mm^4)
        F: Applied force (N)
        E: Young's modulus (MPa)

    Returns:
        float: deflection (m)
               Returns np.nan if calculation error occurs.
    """

    deflection = F * L**3 / (3 * E * 10**6 * I)  # deflection in m
    return deflection


def cb_stress(y, L, I, F):
    """to compute bending stress in the cantilever beam example"""

    try:
        stress = F * L * y / I / 1000  # stress in MPa
    except:
        stress = np.nan

    return stress
